#!/usr/bin/env python
"""This module has tests for the image_records module."""

# Copyright 2022-2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from datetime import datetime, timedelta, timezone
import unittest

from vipersci.pds.pid import VISID
from vipersci.vis.db import image_records as trp
from vipersci.vis.db.image_requests import ImageRequest  # noqa
from vipersci.vis.db.image_stats import ImageStats  # noqa
from vipersci.vis.db.image_tags import ImageTag, taglist  # noqa
from vipersci.vis.db.junc_image_pano import JuncImagePano  # noqa
from vipersci.vis.db.junc_image_record_tags import JuncImageRecordTag  # noqa
from vipersci.vis.db.junc_image_req_ldst import JuncImageRequestLDST  # noqa
from vipersci.vis.db.ldst import LDST  # noqa
from vipersci.vis.db.light_records import LightRecord  # noqa
from vipersci.vis.db.pano_records import PanoRecord  # noqa
from vipersci.vis.db.ptu_records import PanRecord, TiltRecord  # noqa


class TestImageType(unittest.TestCase):
    def test_init(self):
        self.assertEqual(trp.ImageType.LOSSLESS_ICER_IMAGE, trp.ImageType(1))

    def test_single_flags(self):
        self.assertRaises(ValueError, trp.ImageType, 9)

    def test_not_member(self):
        self.assertRaises(ValueError, trp.ImageType, 1000)


class TestProcessingStage(unittest.TestCase):
    def test_init(self):
        self.assertEqual(trp.ProcessingStage.FLATFIELD, trp.ProcessingStage(2))

    def test_combintation(self):
        self.assertEqual(
            trp.ProcessingStage.FLATFIELD | trp.ProcessingStage.LINEARIZATION,
            trp.ProcessingStage(10),
        )

    def test_not_member(self):
        self.assertRaises(ValueError, trp.ProcessingStage, 15)


class TestImageRecord(unittest.TestCase):
    def setUp(self):
        self.startUTC = datetime(2022, 1, 27, 0, 0, 0, tzinfo=timezone.utc)
        self.d = dict(
            adc_gain=63,
            auto_exposure=False,
            bad_pixel_table_id=0,
            capture_id=0,
            exposure_duration=111,
            file_creation_datetime=datetime.now(timezone.utc),
            file_path="/path/to/dummy",
            light_on_hap=False,
            light_on_has=False,
            light_on_hcp=False,
            light_on_hcs=False,
            light_on_hfp=False,
            light_on_hfs=False,
            light_on_nl=False,
            light_on_nr=False,
            icer_byte_quota=493448,
            image_id=0,
            instrument_name="NavCam Left",
            instrument_temperature=128,
            lines=2048,
            lobt=self.startUTC.timestamp(),
            md5_checksum="dummychecksum",
            minLoss=0,
            mission_phase="Test",
            offset=16324,
            # onboard_compression_ratio=5,
            onboard_compression_type="ICER",
            output_image_mask=8,
            output_image_type="?",
            padding=0,
            pga_gain=0,
            processing_info=0,
            purpose="Engineering",
            samples=2048,
            slog=False,
            stereo=False,
            voltage_ramp=109,
        )
        self.extras = dict(foo="bar")

    def test_init(self):
        rp = trp.ImageRecord(**self.d)
        self.assertEqual("220127-000000-ncl-c", str(rp.product_id))

        d = self.d.copy()
        d.update(self.extras)
        rpl = trp.ImageRecord(**d)
        self.assertEqual("220127-000000-ncl-c", str(rpl.product_id))

        d = self.d.copy()
        d["capture_id"] = 65537
        ir_ci = trp.ImageRecord(**d)
        self.assertEqual(1, ir_ci.waypoint_id)
        self.assertEqual(1, ir_ci.unique_capture_id)

    def test_init_slog(self):
        d_slog = {
            "adcGain": 0,
            "autoExposure": 0,
            "cameraId": 0,
            "captureId": 1,
            "exposureTime": 511,
            "imageDepth": 2,
            "imageHeight": 2048,
            "imageWidth": 2048,
            "imageId": 0,
            "immediateDownloadInfo": 24,
            "temperature": 0,
            "lobt": 1698350400,
            "offset": 0,
            "outputImageMask": 16,
            "padding": 0,
            "pgaGain": 1.0,
            "processingInfo": 26,
            "product_id": "231026-200000-ncl-s",
            "stereo": 1,
            "voltageRamp": 0,
            "yamcs_generation_time": "2023-10-26T20:00:00Z",
            "yamcs_reception_time": "2023-10-26T20:03:00Z",
            "yamcs_name": "/ViperGround/Images/ImageData/Navcam_left_slog",
        }
        ir_slog = trp.ImageRecord(**d_slog)
        self.assertEqual("NavCam Left", ir_slog.instrument_name)

        d2_slog = d_slog.copy()
        del d2_slog["outputImageMask"]
        trp.ImageRecord(**d2_slog)

        d3_slog = d_slog.copy()
        del d3_slog["product_id"]
        del d3_slog["outputImageMask"]
        trp.ImageRecord(**d3_slog)

        err_slog = d_slog.copy()
        err_slog["product_id"] = "231026-200000-ncl-z"  # bad compression letter
        err_slog["outputImageMask"] = None
        self.assertRaises(ValueError, trp.ImageRecord, **err_slog)

        # for k in dir(rp):
        #     if k.startswith(("_", "validate_")):
        #         continue

        #     print(f"{k}: {getattr(rp, k)}")

    def test_init_errors(self):
        d = self.d.copy()
        d["start_time"] = self.startUTC + timedelta(hours=1)
        self.assertRaises(ValueError, trp.ImageRecord, **d)

        d = self.d.copy()
        del d["instrument_name"]
        self.assertRaises(ValueError, trp.ImageRecord, **d)

        d = self.d.copy()
        d["product_id"] = "220127-010000-ncl-b"
        self.assertRaises(ValueError, trp.ImageRecord, **d)

        d = self.d.copy()
        d["product_id"] = "220127-000000-ncr-b"
        self.assertRaises(ValueError, trp.ImageRecord, **d)

        d = self.d.copy()
        d["product_id"] = "220127-000000-ncl-b"
        d["onboard_compression_ratio"] = 999
        self.assertRaises(ValueError, trp.ImageRecord, **d)

        d = self.d.copy()
        del d["lobt"]
        d["start_time"] = self.startUTC  # correct
        d["product_id"] = "220127-000001-ncl-c"  # pid.datetime incorrect
        self.assertRaises(ValueError, trp.ImageRecord, **d)

        d = self.d.copy()
        d["cameraId"] = 1
        self.assertWarns(UserWarning, trp.ImageRecord, **d)

        d = self.d.copy()
        d["product_id"] = "220127-000001-ncl-s"  # pid.compression incorrect
        del d["output_image_mask"]
        del d["processing_info"]
        self.assertRaises(ValueError, trp.ImageRecord, **d)

        d = self.d.copy()
        del d["output_image_mask"]
        del d["icer_byte_quota"]
        self.assertRaises(ValueError, trp.ImageRecord, **d)

        d = self.d.copy()
        d["processing_info"] = 99
        self.assertWarns(UserWarning, trp.ImageRecord, **d)

    # Commented out while this exception has been converted to a warning until we
    # sort out the Yamcs parameter.
    # def test_mcam_id(self):
    #     rp = trp.RawProduct(**self.d)
    #     self.assertRaises(ValueError, setattr, rp, "mcam_id", 5)

    def test_product_id(self):
        rp = trp.ImageRecord(**self.d)
        self.assertRaises(NotImplementedError, setattr, rp, "product_id", "dummy")

    # def test_purpose(self):
    #     rp = trp.ImageRecord(**self.d)
    #     self.assertRaises(ValueError, setattr, rp, "purpose", "dummy")

    def test_lt(self):
        ir1 = trp.ImageRecord(**self.d)
        v = VISID("230127-000000-ncl-a")
        ir2 = trp.ImageRecord(
            product_id=str(v), start_time=v.datetime(), exposure_duration=111
        )
        self.assertTrue(ir1 < ir2)

        self.assertEqual(NotImplemented, ir1.__lt__("not an ImageRecord"))

    def test_update(self):
        rp = trp.ImageRecord(**self.d)
        k = "foo"
        self.assertTrue(k not in rp.labelmeta)

        rp.update(self.extras)
        self.assertTrue(k in rp.labelmeta)

        rp.update({"file_md5_checksum": "foo"})
        self.assertEqual(rp.file_md5_checksum, "foo")

    # def test_labeldict(self):
    #     din = self.d
    #     din.update(self.extras)
    #     rp = trp.ImageRecord(**din)
    #     d = rp.label_dict()
    #     self.assertEqual(d["samples"], rp.samples)

    def test_synonym(self):
        din = self.d
        del din["exposure_duration"]
        din["exposureTime"] = 400
        del din["samples"]
        din["imageWidth"] = 4
        rp = trp.ImageRecord(**din)
        self.assertEqual(rp.exposure_duration, 400)
        rp.exposureTime = 500
        self.assertEqual(rp.exposure_duration, 500)

    def test_fromyamcs(self):
        name = "/ViperGround/Images/ImageData/Navcam_left_slog"
        reception_time = datetime.now(timezone.utc)
        generation_time = reception_time - timedelta(minutes=1)
        d = {
            "adcGain": 0,
            "autoExposure": 0,
            "cameraId": 0,
            "captureId": 1,
            "exposureTime": 111,
            "haz1LightState": "OFF",
            "haz2LightState": "OFF",
            "haz3LightState": "OFF",
            "haz4LightState": "OFF",
            "haz5LightState": "OFF",
            "haz6LightState": "OFF",
            "imageDepth": 1,
            "imageHeight": 2048,
            "imageId": 0,
            "imageWidth": 2048,
            "immediateDownloadInfo": 10,
            "lobt": 1700921056,
            "navLeftLightState": "ON",
            "navRightLightState": "ON",
            "offset": 0,
            "outputImageMask": 16,
            "outputImageType": "JBIG2_IMAGE",
            "padding": 0,
            "ppaGain": 0,
            "processingInfo": 26,
            "stereo": 1,
            "temperature": 0,
            "voltageRamp": 0,
        }
        rp = trp.ImageRecord(
            yamcs_name=name,
            yamcs_generation_time=generation_time,
            yamcs_reception_time=reception_time,
            **d,
            onboard_compression_ratio=16
        )
        self.assertEqual(rp.yamcs_name, name)
        self.assertEqual(rp.samples, d["imageWidth"])

    def test_fromxml(self):
        t = """<?xml version="1.0" encoding="UTF-8"?>
<?xml-model href="http://pds.nasa.gov/pds4/pds/v1/PDS4_PDS_1I00.sch" schematypens="http://purl.oclc.org/dsdl/schematron"?>
<?xml-model href="http://pds.nasa.gov/pds4/disp/v1/PDS4_DISP_1I00_1510.sch" schematypens="http://purl.oclc.org/dsdl/schematron"?>
<?xml-model href="http://pds.nasa.gov/pds4/img/v1/PDS4_IMG_1I00_1860.sch" schematypens="http://purl.oclc.org/dsdl/schematron"?>
<?xml-model href="http://pds.nasa.gov/pds4/msn/v1/PDS4_MSN_1I00_1300.sch" schematypens="http://purl.oclc.org/dsdl/schematron"?>
<?xml-model href="http://pds.nasa.gov/pds4/proc/v1/PDS4_PROC_1I00_1210.sch" schematypens="http://purl.oclc.org/dsdl/schematron"?>

<Product_Observational
    xmlns="http://pds.nasa.gov/pds4/pds/v1"
    xmlns:disp="http://pds.nasa.gov/pds4/disp/v1"
    xmlns:img="http://pds.nasa.gov/pds4/img/v1"
    xmlns:msn="http://pds.nasa.gov/pds4/msn/v1"
    xmlns:proc="http://pds.nasa.gov/pds4/proc/v1"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:schemaLocation="
    http://pds.nasa.gov/pds4/pds/v1 http://pds.nasa.gov/pds4/pds/v1/PDS4_PDS_1I00.xsd
    http://pds.nasa.gov/pds4/disp/v1 http://pds.nasa.gov/pds4/disp/v1/PDS4_DISP_1I00_1510.xsd
    http://pds.nasa.gov/pds4/img/v1 http://pds.nasa.gov/pds4/img/v1/PDS4_IMG_1I00_1860.xsd
    http://pds.nasa.gov/pds4/msn/v1 http://pds.nasa.gov/pds4/msn/v1/PDS4_MSN_1I00_1300.xsd
    http://pds.nasa.gov/pds4/proc/v1 http://pds.nasa.gov/pds4/proc/v1/PDS4_PROC_1I00_1210.xsd
">
  <Identification_Area>
    <logical_identifier>urn:nasa:pds:viper_vis:raw:231125-143859-ncl-d</logical_identifier>
    <version_id>0.1</version_id>
    <title>VIPER Visible Imaging System NavCam Left image - 231125-143859-ncl-d</title>
    <information_model_version>1.18.0.0</information_model_version>
    <product_class>Product_Observational</product_class>
    <Modification_History>
      <Modification_Detail>
        <modification_date>2022-10-19</modification_date>
        <version_id>0.1</version_id>
        <description>Illegal version number for testing</description>
      </Modification_Detail>
    </Modification_History>
  </Identification_Area>
  <Observation_Area>
    <Time_Coordinates>
      <start_date_time>2023-11-25T14:38:59Z</start_date_time>
      <stop_date_time>2023-11-25T14:38:59.000111Z</stop_date_time>
    </Time_Coordinates>
    <Primary_Result_Summary>
      <purpose>Navigation</purpose>
      <processing_level>Raw</processing_level>
    </Primary_Result_Summary>
    <Investigation_Area>
      <name>VIPER</name>
      <type>Mission</type>
      <Internal_Reference>
        <lid_reference>urn:nasa:pds:viper</lid_reference>
        <reference_type>data_to_investigation</reference_type>
      </Internal_Reference>
    </Investigation_Area>
    <Observing_System>
      <Observing_System_Component>
        <name>VIPER</name>
        <type>Host</type>
        <Internal_Reference>
            <lid_reference>urn:nasa:pds:context:instrument_host:spacecraft.viper</lid_reference>
            <reference_type>is_instrument_host</reference_type>
        </Internal_Reference>
      </Observing_System_Component>
      <Observing_System_Component>
        <name>NavCam Left</name>
        <type>Instrument</type>
        <Internal_Reference>
            <lid_reference>urn:nasa:pds:context:instrument_host:spacecraft.viper.navcam_left</lid_reference>
            <reference_type>is_instrument</reference_type>
        </Internal_Reference>
      </Observing_System_Component>
    </Observing_System>
    <Target_Identification>
      <name>Moon</name>
      <type>Satellite</type>
      <Internal_Reference>
        <lid_reference>urn:nasa:pds:context:target:satellite.earth.moon</lid_reference>
        <reference_type>data_to_target</reference_type>
      </Internal_Reference>
    </Target_Identification>
    <Discipline_Area>
        <disp:Display_Settings>
            <Local_Internal_Reference>
                <local_identifier_reference>image2d</local_identifier_reference>
                <local_reference_type>display_settings_to_array</local_reference_type>
            </Local_Internal_Reference>
            <disp:Display_Direction>
                <disp:horizontal_display_axis>Sample</disp:horizontal_display_axis>
                <disp:horizontal_display_direction>Left to Right</disp:horizontal_display_direction>
                <disp:vertical_display_axis>Line</disp:vertical_display_axis>
                <disp:vertical_display_direction>Top to Bottom</disp:vertical_display_direction>
            </disp:Display_Direction>
        </disp:Display_Settings>
        <img:Imaging>
           <Local_Internal_Reference>
               <local_identifier_reference>image2d</local_identifier_reference>
               <local_reference_type>imaging_parameters_to_image_object</local_reference_type>
           </Local_Internal_Reference>
           <img:Detector>
               <img:first_line>1</img:first_line>
               <img:first_sample>1</img:first_sample>
               <img:lines>2048</img:lines>
               <img:samples>2048</img:samples>
               <img:gain_number>1</img:gain_number>
               <img:analog_offset>0</img:analog_offset>
               <img:bad_pixel_replacement_table_id>0</img:bad_pixel_replacement_table_id>
           </img:Detector>
           <img:Exposure>
               <img:exposure_duration unit="microseconds">111</img:exposure_duration>
               <img:exposure_type>Manual</img:exposure_type>
           </img:Exposure>
            <img:Illumination>
                <img:LED_Illumination_Source>
                    <img:name>NavLight Left</img:name>
                    <img:illumination_state>Off</img:illumination_state>
                    <img:illumination_wavelength unit="nm">453</img:illumination_wavelength>
                </img:LED_Illumination_Source><img:LED_Illumination_Source>
                    <img:name>NavLight Right</img:name>
                    <img:illumination_state>Off</img:illumination_state>
                    <img:illumination_wavelength unit="nm">453</img:illumination_wavelength>
                </img:LED_Illumination_Source><img:LED_Illumination_Source>
                    <img:name>HazLight Aft Port</img:name>
                    <img:illumination_state>Off</img:illumination_state>
                    <img:illumination_wavelength unit="nm">453</img:illumination_wavelength>
                </img:LED_Illumination_Source><img:LED_Illumination_Source>
                    <img:name>HazLight Aft Starboard</img:name>
                    <img:illumination_state>Off</img:illumination_state>
                    <img:illumination_wavelength unit="nm">453</img:illumination_wavelength>
                </img:LED_Illumination_Source><img:LED_Illumination_Source>
                    <img:name>HazLight Center Port</img:name>
                    <img:illumination_state>Off</img:illumination_state>
                    <img:illumination_wavelength unit="nm">453</img:illumination_wavelength>
                </img:LED_Illumination_Source><img:LED_Illumination_Source>
                    <img:name>HazLight Center Starboard</img:name>
                    <img:illumination_state>Off</img:illumination_state>
                    <img:illumination_wavelength unit="nm">453</img:illumination_wavelength>
                </img:LED_Illumination_Source><img:LED_Illumination_Source>
                    <img:name>HazLight Fore Port</img:name>
                    <img:illumination_state>Off</img:illumination_state>
                    <img:illumination_wavelength unit="nm">453</img:illumination_wavelength>
                </img:LED_Illumination_Source><img:LED_Illumination_Source>
                    <img:name>HazLight Fore Starboard</img:name>
                    <img:illumination_state>Off</img:illumination_state>
                    <img:illumination_wavelength unit="nm">453</img:illumination_wavelength>
                </img:LED_Illumination_Source>
            </img:Illumination>
        <img:Onboard_Compression>
            <img:onboard_compression_class>Lossy</img:onboard_compression_class>
            <img:onboard_compression_type>ICER</img:onboard_compression_type>
            <img:onboard_compression_ratio>64</img:onboard_compression_ratio>
        </img:Onboard_Compression>
        <img:Sampling>
            <img:sample_bits>12</img:sample_bits>
            <img:sample_bit_mask>2#0000111111111111</img:sample_bit_mask>
        </img:Sampling>
        <img:Instrument_State>
            <img:Device_Temperatures>
                <img:Device_Temperature>
                    <img:device_name>NavCam Left</img:device_name>
                    <img:temperature_value unit="K">0</img:temperature_value>
                </img:Device_Temperature>
            </img:Device_Temperatures>
        </img:Instrument_State>
        </img:Imaging>
        <msn:Mission_Information>
            <msn:mission_phase_name>TEST</msn:mission_phase_name>
        </msn:Mission_Information>
        <proc:Processing_Information>
            <Local_Internal_Reference>
                <local_identifier_reference>image2d</local_identifier_reference>
                <local_reference_type>processing_information_to_data_object</local_reference_type>
            </Local_Internal_Reference>
            <proc:Process>
                <proc:process_owner_institution_name>VIPER Visible Imaging System Team,
                NASA Ames Research Center</proc:process_owner_institution_name>
                <proc:Software>
                    <proc:name>vipersci</proc:name>
                    <proc:software_version_id>0.1.0</proc:software_version_id>
                    <proc:software_type>Python</proc:software_type>
                    <proc:Software_Program>
                        <proc:name>vipersci.vis.pds.create_raw</proc:name>
                    </proc:Software_Program>
                </proc:Software>
            </proc:Process>
        </proc:Processing_Information>
    </Discipline_Area>
  </Observation_Area>
  <File_Area_Observational>
    <File>
      <file_name>231125-143859-ncl-d.tif</file_name>
      <creation_date_time>2022-10-19T17:27:04.097587Z</creation_date_time>
    </File>
    <Array_2D_Image>
        <local_identifier>image2d</local_identifier>
        <md5_checksum>8c708f5745ad2b6d9bac6036062bbd31</md5_checksum>
        <offset unit="byte">256</offset>
        <axes>2</axes>
        <axis_index_order>Last Index Fastest</axis_index_order>
        <Element_Array>
            <data_type>UnsignedLSB2</data_type>
            <unit>DN</unit>
        </Element_Array>
        <Axis_Array>
            <axis_name>Line</axis_name>
            <elements>2048</elements>
            <sequence_number>1</sequence_number>
        </Axis_Array>
        <Axis_Array>
            <axis_name>Sample</axis_name>
            <elements>2048</elements>
            <sequence_number>2</sequence_number>
        </Axis_Array>
    </Array_2D_Image>
  </File_Area_Observational>
</Product_Observational>
        """  # noqa: E501
        rp = trp.ImageRecord.from_xml(t.encode())
        self.assertEqual("231125-143859-ncl-d", rp.product_id)

        t_not_viper_vis = t.replace(
            "<logical_identifier>urn:nasa:pds:viper_vis:raw:231125-143859-ncl-d",
            "<logical_identifier>urn:nasa:pds:NOT_viper_vis:raw:231125-143859-ncl-d",
        )
        self.assertRaises(
            ValueError, trp.ImageRecord.from_xml, t_not_viper_vis.encode()
        )

        t_not_raw = t.replace(
            "<logical_identifier>urn:nasa:pds:viper_vis:raw:231125-143859-ncl-d",
            "<logical_identifier>urn:nasa:pds:viper_vis:NOT_raw:231125-143859-ncl-d",
        )
        self.assertRaises(ValueError, trp.ImageRecord.from_xml, t_not_raw.encode())



#!/usr/bin/env python
"""This module has tests for the vis.create_mmgis_pano functions."""

# Copyright 2023, vipersci developers.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from argparse import ArgumentParser
from pathlib import Path
import unittest
from unittest.mock import create_autospec, mock_open, patch

import numpy as np

from vipersci.pds import pid as pds
from vipersci.vis import create_mmgis_pano as cp


class TestCLI(unittest.TestCase):
    def test_arg_parser(self):
        p = cp.arg_parser()
        self.assertIsInstance(p, ArgumentParser)
        # self.assertRaises(SystemExit, p.parse_args)
        d = vars(
            p.parse_args(
                ["--dburl", "db://foo:username@host/db", "product_id_goes_here"]
            )
        )
        self.assertIn("dburl", d)
        self.assertIn("input", d)

    # def test_main(self):
    #     pa_ret_val = cp.arg_parser().parse_args(
    #         ["231126-000000-ncl-s.dummy", "231126-000000-ncr-s.dummy"]
    #     )
    #     with patch("vipersci.vis.create_pano.arg_parser") as parser, patch(
    #         "vipersci.vis.create_pano.create"
    #     ) as m_create:
    #         parser.return_value.parse_args.return_value = pa_ret_val
    #         cp.main()
    #         m_create.assert_called_once()

    #     pa2_ret_val = cp.arg_parser().parse_args(
    #         [
    #             "--dburl",
    #             "db://foo:username@host/db",
    #             "231126-000000-ncl-s.dummy",
    #             "231126-000000-ncr-s.dummy",
    #         ]
    #     )
    #     session_engine_mock = create_autospec(Session)
    #     session_mock = create_autospec(Session)
    #     session_engine_mock.__enter__ = Mock(return_value=session_mock)
    #     with patch("vipersci.vis.create_pano.arg_parser") as parser, patch(
    #         "vipersci.vis.create_pano.create"
    #     ) as m_create, patch("vipersci.vis.create_pano.create_engine"), patch(
    #         "vipersci.vis.create_pano.Session", return_value=session_engine_mock
    #     ):
    #         parser.return_value.parse_args.return_value = pa2_ret_val
    #         cp.main()
    #         m_create.assert_called_once()
    #         session_engine_mock.__enter__.assert_called_once()
    #         session_mock.commit.assert_called_once()


class TestOther(unittest.TestCase):
    def test_longname(self):
        ln = cp.longname(Path("path/to/240225-000000-pan.dummy"))
        self.assertEqual(ln, "2024-02-25T00-00-00-pan")

    def test_mmgis_data(self):
        pd = {
            "rover_pan_max": 30,
            "rover_pan_min": 5,
            "samples": 5000,
            "lines": 2048,
            "rover_tilt_max": 20,
            "rover_tilt_min": -10,
            "product_id": "240225-000000-pan",
        }
        d = cp.mmgis_data(pd)
        self.assertEqual(d["azmax"], pd["rover_pan_max"])
        self.assertEqual(d["azmin"], pd["rover_pan_min"])
        self.assertEqual(d["columns"], pd["samples"])
        self.assertEqual(d["rows"], pd["lines"])
        self.assertEqual(d["elmax"], pd["rover_tilt_max"])
        self.assertEqual(d["elmin"], pd["rover_tilt_min"])
        self.assertEqual(d["name"], pd["product_id"])
        self.assertTrue(d["isPanoramic"])


class TestCreate(unittest.TestCase):

    @patch("vipersci.vis.create_mmgis_pano.json.dump")
    @patch("vipersci.vis.create_mmgis_pano.imread")
    @patch("vipersci.vis.create_mmgis_pano.imsave")
    @patch("vipersci.vis.create_mmgis_pano.equalize_adapthist")
    @patch(
        "vipersci.vis.create_mmgis_pano.rescale_intensity",
        return_value=create_autospec(np.ndarray),
    )
    @patch("vipersci.vis.create_mmgis_pano.resize")
    @patch("vipersci.vis.create_mmgis_pano.mmgis_data")
    def test_nodb(
        self,
        m_mmgis_data,
        m_resize,
        m_rescale_int,
        m_eq_adapt,
        m_imsave,
        m_imread,
        m_dump,
    ):
        self.assertRaises(ValueError, cp.create, pds.PanoID("240225-000000-pan"))
        self.assertRaises(ValueError, cp.create, ("Not a recognized", "type."))

        path_mock = create_autospec(Path)
        info = {"file_path": "240225-000000-pan.tif"}
        with patch(
            "vipersci.vis.create_mmgis_pano.json.load", return_value=info
        ) as m_load, patch(
            "vipersci.vis.create_mmgis_pano.open", mock_open()
        ) as m_open:
            cp.create(path_mock)
            self.assertEqual(m_open.call_count, 2)
            self.assertEqual(m_resize.call_count, 1)
            self.assertEqual(m_rescale_int.call_count, 2)
            self.assertEqual(m_eq_adapt.call_count, 1)
            self.assertEqual(m_imsave.call_count, 2)
            self.assertEqual(m_load.call_count, 1)
            self.assertEqual(m_dump.call_count, 1)
            m_imread.assert_called_once_with(str(Path(info["file_path"])))
            m_mmgis_data.assert_called_once_with(info, 0.0)



#!/usr/bin/env python
"""This module has tests for the vis.pds.create_raw functions."""

# Copyright 2022-2023, vipersci developers.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from argparse import ArgumentParser
from datetime import datetime, timezone
from pathlib import Path
import unittest
from unittest.mock import create_autospec, Mock, mock_open, patch

import numpy as np
import numpy.testing as npt
from PIL import Image
from sqlalchemy.orm import Session

from vipersci.pds import pid as pds
from vipersci.vis.db.image_records import ImageRecord
from vipersci.vis import create_image as ci


class TestCLI(unittest.TestCase):
    def test_arg_parser(self):
        p = ci.arg_parser()
        self.assertIsInstance(p, ArgumentParser)
        self.assertRaises(SystemExit, p.parse_args)
        d = vars(p.parse_args(["-o", "output_directory", "dummy.json"]))
        self.assertIn("output_dir", d)
        self.assertIn("input", d)

    def test_main(self):
        pa_ret_val = ci.arg_parser().parse_args(
            [
                "input.json",
            ]
        )
        with patch("vipersci.vis.create_image.arg_parser") as parser, patch(
            "vipersci.vis.create_image.create"
        ) as m_create, patch(
            "vipersci.vis.create_image.open", mock_open(read_data='{"json": "dummy"}')
        ):
            parser.return_value.parse_args.return_value = pa_ret_val
            ci.main()
            m_create.assert_called_once()

    def test_main_db(self):
        pa2_ret_val = ci.arg_parser().parse_args(
            [
                "--dburl",
                "db://foo:username@host/db",
                "--tiff",
                "dummy.tif",
                "dummy.json",
            ]
        )
        session_engine_mock = create_autospec(Session)
        session_mock = create_autospec(Session)
        session_engine_mock.__enter__ = Mock(return_value=session_mock)
        with patch("vipersci.vis.create_image.arg_parser") as parser, patch(
            "vipersci.vis.create_image.create"
        ) as m_create, patch("vipersci.vis.create_image.create_engine"), patch(
            "vipersci.vis.create_image.Session", return_value=session_engine_mock
        ), patch(
            "vipersci.vis.create_image.open", mock_open(read_data='{"json": "dummy"}')
        ):
            parser.return_value.parse_args.return_value = pa2_ret_val
            ci.main()
            m_create.assert_called_once()
            session_engine_mock.__enter__.assert_called_once()
            session_mock.commit.assert_called_once()

    def test_main_image(self):

        pa_ret_val = ci.arg_parser().parse_args(["--image", "dummy.png", "input.json"])
        with patch("vipersci.vis.create_image.arg_parser") as parser, patch(
            "vipersci.vis.create_image.create"
        ) as m_create, patch(
            "vipersci.vis.create_image.open", mock_open(read_data='{"json": "dummy"}')
        ), patch(
            "vipersci.vis.create_image.imread"
        ):
            parser.return_value.parse_args.return_value = pa_ret_val
            ci.main()
            m_create.assert_called_once()


class TestBitDepth(unittest.TestCase):
    def test_check_bit_depth(self):
        pid = pds.VISID("231125-143859-ncl-a")
        self.assertIsNone(ci.check_bit_depth(pid, 16))
        self.assertIsNone(ci.check_bit_depth(pid, "UnsignedLSB2"))
        self.assertIsNone(ci.check_bit_depth(pid, np.uint16(5).dtype))

        self.assertRaises(ValueError, ci.check_bit_depth, pid, None)
        self.assertRaises(ValueError, ci.check_bit_depth, pid, 8)

        pids = pds.VISID("231125-143859-ncl-s")
        self.assertIsNone(ci.check_bit_depth(pids, np.uint8(5).dtype))
        self.assertIsNone(ci.check_bit_depth(pids, "Byte"))
        self.assertRaises(ValueError, ci.check_bit_depth, pids, 16)


class TestCreate(unittest.TestCase):
    def test_create(self):
        d = {"meta": "data"}
        with patch("vipersci.vis.create_image.make_image_record") as m_mir:
            ci.create(d, json=False)
            m_mir.assert_called_once_with(d, None, Path.cwd(), "tif")

        with patch("vipersci.vis.create_image.make_image_record") as m_mir, patch(
            "vipersci.vis.create_image.write_json"
        ) as m_wj:
            session_mock = create_autospec(Session)
            ci.create(d, session=session_mock)
            m_wj.assert_called_once()


class TestMakeImage(unittest.TestCase):
    def setUp(self) -> None:
        self.startUTC = datetime(2022, 1, 27, 0, 0, 0, tzinfo=timezone.utc)
        self.d = {
            "adcGain": 0,
            "autoExposure": 0,
            "byteQuota": 1677,
            "cameraId": 0,
            "captureId": 1,
            "exposureTime": 111,
            "imageDepth": 1,
            "imageHeight": 2048,
            "imageId": 0,
            "imageWidth": 2048,
            "immediateDownloadInfo": 10,
            "instrument_name": "NavCam Left",
            "lobt": 1700921056,
            "offset": 0,
            "outputImageMask": 8,
            "outputImageType": "JBIG2_IMAGE",
            "padding": 0,
            "ppaGain": 0,
            "processingInfo": 0,
            "slog": False,
            "stereo": 1,
            "temperature": 0,
            "voltageRamp": 0,
        }

    def test_no_image(self):
        rp = ci.make_image_record(self.d, None)
        self.assertIsInstance(rp, ImageRecord)

    @patch("vipersci.vis.create_image.write_tiff")
    @patch(
        "vipersci.vis.create_image.tif_info",
        return_value={
            "file_byte_offset": 10,
            "file_creation_datetime": datetime.fromtimestamp(1700921056, timezone.utc),
            "file_data_type": "UnsignedMSB2",
            "file_md5_checksum": "md5",
            "file_path": "dummy.tif",
            "lines": 2048,
            "samples": 2048,
        },
    )
    def test_image(self, mock_tif_info, mock_write_tiff):
        image = np.array([[5, 5], [5, 5]], dtype=np.uint16)
        rp = ci.make_image_record(self.d, image, Path("outdir/"))
        self.assertIsInstance(rp, ImageRecord)

        prp = ci.make_image_record(self.d, Path("dummy.tif"), Path("outdir/"))
        self.assertIsInstance(prp, ImageRecord)

        bad_d = self.d
        bad_d["imageWidth"] = 2
        self.assertRaises(
            ValueError, ci.make_image_record, bad_d, image, Path("outdir/")
        )

    @patch("vipersci.vis.create_image.write_png")
    @patch(
        "vipersci.vis.create_image.png_info",
        return_value={
            "file_creation_datetime": datetime.fromtimestamp(1700921056, timezone.utc),
            "file_data_type": "UnsignedMSB2",
            "file_md5_checksum": "md5",
            "file_path": "dummy.tif",
            "lines": 2048,
            "samples": 2048,
        },
    )
    def test_png(self, mock_png_info, mock_write_png):
        image = np.array([[5, 5], [5, 5]], dtype=np.uint16)
        rp = ci.make_image_record(self.d, image, Path("outdir/"), "png")
        self.assertIsInstance(rp, ImageRecord)

        prp = ci.make_image_record(self.d, Path("dummy.png"), Path("outdir/"), "png")
        self.assertIsInstance(prp, ImageRecord)

    def test_bad_imgtype(self):
        self.assertRaises(
            ValueError,
            ci.make_image_record,
            self.d,
            Path("dummy.foo"),
            Path("outdir/"),
            "tif",
        )
        self.assertRaises(
            ValueError,
            ci.make_image_record,
            self.d,
            np.array([[5, 5], [5, 5]], dtype=np.uint16),
            Path("outdir/"),
            "foo",
        )


class TestJSON(unittest.TestCase):
    @patch("vipersci.vis.create_image.json.dump")
    def test_write_json(self, m_dump):
        mock_path = create_autospec(Path)
        mock_path.name = "dummy_dir"
        d = {"key": "value", "key2": "value2", "product_id": "dummy_pid"}
        ci.write_json(d, mock_path)
        m_dump.called_once()


class TestPNG(unittest.TestCase):

    @patch("vipersci.vis.create_image.file_info", return_value={"file_path": "name"})
    def test_png_info(self, mock_fi):
        mock_im = create_autospec(Image)
        mock_im.size = [10, 20]
        with patch("vipersci.vis.create_image.Image.open", return_value=mock_im):
            d = ci.png_info(Path("dummy.png"))
            self.assertDictEqual(d, {"lines": 20, "samples": 10, "file_path": "name"})

    @patch("vipersci.vis.create_image.imsave")
    def test_write_png(self, mock_imsave):
        pid = pds.VISID("231125-143859-ncl-a")
        image = np.array([[5, 5], [5, 5]], dtype=np.uint16)
        outpath = ci.write_png(pid, image, Path("dummy/"))
        self.assertEqual(outpath, Path("dummy/231125-143859-ncl-a.png"))

        self.assertEqual(mock_imsave.call_args.args[0], str(outpath))
        npt.assert_array_equal(mock_imsave.call_args.args[1], image)
        self.assertEqual(mock_imsave.call_args.kwargs["check_contrast"], False)
        self.assertIn("pnginfo", mock_imsave.call_args.kwargs)

        image_bool = np.array([[5, 5], [5, 5]], dtype=np.bool_)
        ci.write_png(pds.VISID("231125-143859-ncl-s"), image_bool, Path("dummy/"))

        image_32 = np.array([[5, 5], [5, 5]], dtype=np.int32)
        ci.write_png(pid, image_32, Path("dummy/"))


class TestTIFF(unittest.TestCase):
    @patch("vipersci.vis.create_image.datetime", wraps=datetime)
    def test_tif_info(self, mock_datetime):
        info = {
            "ifds": [
                {
                    "tags": {
                        273: {"data": (10,)},
                        258: {"data": (16,)},
                        256: {"data": (2048,)},
                        257: {"data": (2048,)},
                    }
                },
            ],
            "bigEndian": "MSB",
        }
        mock_datetime.fromtimestamp.return_value = datetime.fromtimestamp(
            1700921056, timezone.utc
        )

        mock_path = create_autospec(Path)
        mock_path.name = "dummy.tif"

        with patch("vipersci.vis.create_image.util.md5", return_value="hex"):
            with patch("vipersci.vis.create_image.read_tiff", return_value=info):
                d = ci.tif_info(mock_path)

                self.assertEqual(d["file_byte_offset"], 10)
                self.assertEqual(d["file_data_type"], "UnsignedMSB2")

        info2 = {
            "ifds": [
                {
                    "tags": {
                        273: {"data": (10,)},
                        258: {"data": (8,)},
                        256: {"data": (2048,)},
                        257: {"data": (2048,)},
                    }
                },
            ],
            "bigEndian": "MSB",
        }

        with patch("vipersci.vis.create_image.util.md5", return_value="hex"):
            with patch("vipersci.vis.create_image.read_tiff", return_value=info2):
                d = ci.tif_info(mock_path)

                self.assertEqual(d["file_data_type"], "UnsignedByte")

    @patch("vipersci.vis.create_image.imsave")
    def test_write_tiff(self, mock_imsave):
        pid = pds.VISID("231125-143859-ncl-a")
        image = np.array([[5, 5], [5, 5]], dtype=np.uint16)
        outpath = ci.write_tiff(pid, image, Path("dummy/"))
        self.assertEqual(outpath, Path("dummy/231125-143859-ncl-a.tif"))
        mock_imsave.assert_called_once_with(
            str(outpath),
            image,
            check_contrast=False,
            description="VIPER NavCam Left 231125-143859-ncl-a",
            metadata=None,
        )

        image_bool = np.array([[5, 5], [5, 5]], dtype=np.bool_)
        ci.write_tiff(pds.VISID("231125-143859-ncl-s"), image_bool, Path("dummy/"))

        image_32 = np.array([[5, 5], [5, 5]], dtype=np.int32)
        ci.write_tiff(pid, image_32, Path("dummy/"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the `vis/pds` module."""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from pathlib import Path
import unittest
from unittest.mock import create_autospec, patch

from genshi.template import Template

from vipersci.vis import pds


class TestVersion(unittest.TestCase):
    def test_version_info(self):
        d = pds.version_info()
        self.assertEqual(99.99, d["vid"])


class TestXML(unittest.TestCase):
    def test_write_xml(self):
        product = {
            "dummy": "product",
            "product_id": "dummy_pid",
        }
        m_outdir = create_autospec(Path)
        m_tmpl = create_autospec(Template)
        with patch(
            "vipersci.vis.pds.MarkupTemplate", return_value=m_tmpl
        ) as m_markup, patch("vipersci.vis.pds.resources.read_text") as m_read_text:
            pds.write_xml(product, "raw-template.xml", m_outdir)

            m_read_text.assert_called_once()
            m_markup.assert_called_once()
            m_tmpl.generate.assert_called_once()

    def test_write_xml_custom_template(self):
        product = {
            "dummy": "product",
            "product_id": "dummy_pid",
        }

        path_mock = create_autospec(Path)
        path_mock.exists.return_value = True

        m_outdir = create_autospec(Path)
        m_tmpl = create_autospec(Template)

        with patch("vipersci.vis.pds.Path", return_value=path_mock), patch(
            "vipersci.vis.pds.MarkupTemplate", return_value=m_tmpl
        ) as m_markup:
            pds.write_xml(product, "this_exists.xml", m_outdir)

            path_mock.read_text.assert_called_once()
            m_markup.assert_called_once()
            m_tmpl.generate.assert_called_once()



"""Generically takes a Genshi XML template and a JSON or YAML file to complete
the template.
"""

# Copyright 2021-2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import json
import logging
from pathlib import Path
import yaml

from vipersci.pds.labelmaker import write_xml
from vipersci import util

logger = logging.getLogger(__name__)


def add_parser(subparsers):
    parser = subparsers.add_parser("generic", help=__doc__)
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-j", "--json", type=Path, help="Path to .json file to load.")
    group.add_argument("-y", "--yaml", type=Path, help="Path to .yml file to load.")
    parser.add_argument("template", type=Path, help="Genshi XML file template.")
    parser.add_argument("output", type=Path, help="Output XML label.")
    parser.set_defaults(func=main)
    return parser


def main(args):
    util.set_logger(args.verbose)

    info = {}
    if args.json is not None:
        info = json.loads(args.json.read_text())
    elif args.yaml is not None:
        info = yaml.safe_load(args.yaml.read_text())

    write_xml(info, args.output, args.template)

    return



#!/usr/bin/env python
"""This module has tests for the vis.pds.labelmaker.generic functions."""

# Copyright 2023, vipersci developers.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from argparse import ArgumentParser
from pathlib import Path
import unittest
from unittest.mock import create_autospec, patch

import vipersci.pds.labelmaker.generic as gen
from vipersci import util


class TestParser(unittest.TestCase):
    def test_add_parser(self):
        parser = ArgumentParser()
        subparsers = parser.add_subparsers()
        gen.add_parser(subparsers)

        d = vars(
            parser.parse_args(
                ["generic", "--json", "dumb.json", "template.xml", "out.xml"]
            )
        )
        self.assertIn("json", d)
        self.assertIn("template", d)
        self.assertIn("output", d)

    def test_main(self):
        parser = ArgumentParser(parents=[util.parent_parser()])
        subparsers = parser.add_subparsers()
        gen.add_parser(subparsers)

        args = parser.parse_args(
            ["generic", "--json", "dumb.json", "template.xml", "out.xml"]
        )

        args.json = create_autospec(Path)

        with patch(
            "vipersci.pds.labelmaker.generic.json.loads", return_value={"dummy": "dict"}
        ) as m_jsonloads, patch(
            "vipersci.pds.labelmaker.generic.write_xml",
        ) as m_write_xml:
            gen.main(args)

            m_jsonloads.assert_called_once()
            m_write_xml.assert_called_once()



"""Extracts a "publishable" set of bundle and collection files when pointed at
a bundle directory.
"""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
import csv
import logging
from pathlib import Path
from shutil import copy2
import xml.etree.ElementTree as ET

from vipersci.pds.labelmaker import get_lidvidfile
from vipersci.pds.xml import find_text, ns
from vipersci import util

logger = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__, parents=[util.parent_parser()]
    )
    parser.add_argument(
        "source_directory",
        type=Path,
        help="A directory that contains created Bundle files.",
    )
    parser.add_argument(
        "build_directory",
        type=Path,
        help="Path to a directory where the bundle files will be installed.",
    )
    return parser


def main():
    parser = arg_parser()
    args = parser.parse_args()
    util.set_logger(args.verbose)

    args.build_directory.mkdir(exist_ok=True)

    copy2(args.source_directory / "bundle.xml", args.build_directory)

    bundle = ET.fromstring((args.build_directory / "bundle.xml").read_text())
    readme = bundle.find("./pds:File_Area_Text/pds:File/pds:file_name", ns)
    if readme is not None:
        copy2(args.source_directory / readme.text, args.build_directory)

    for bme in bundle.findall(".//pds:Bundle_Member_Entry", ns):
        try:
            lidvid_ref = find_text(bme, "pds:lidvid_reference")
            col_lid, col_vid = lidvid_ref.split("::")
        except ValueError:
            col_lid = find_text(bme, "pds:lid_reference")
            col_vid = None

        if find_text(bme, "pds:member_status") == "Primary":
            col_name = col_lid.split(":")[-1]
            src_col_dir = args.source_directory / col_name
            if src_col_dir.exists() is False:
                raise FileNotFoundError(f"{src_col_dir} does not exist.")

            bld_col_dir = args.build_directory / col_name
            bld_col_dir.mkdir(exist_ok=True)

            col_label_file = f"collection_{col_name}.xml"
            copy2(src_col_dir / col_label_file, bld_col_dir)
            collection = ET.fromstring((bld_col_dir / col_label_file).read_text())
            this_col_lid = find_text(
                collection, "./pds:Identification_Area/pds:logical_identifier"
            )
            if col_lid != this_col_lid:
                raise ValueError(
                    f"The collection lid ({this_col_lid}) does not match the "
                    f"collection lid that the bundle file has ({col_lid})."
                )
            if col_vid is not None:
                this_col_vid = find_text(
                    collection, "./pds:Identification_Area/pds:version_id"
                )
                if this_col_vid != col_vid:
                    raise ValueError(
                        f"The collection vid ({this_col_vid}) for {col_lid} does not "
                        f"match the collection vid that the bundle file "
                        f"has ({col_vid})."
                    )
            inventory = find_text(
                collection, "./pds:File_Area_Inventory/pds:File/pds:file_name"
            )
            copy2(src_col_dir / inventory, bld_col_dir)
            col_lidvids = set()
            with open(str(src_col_dir / inventory), newline="") as csvfile:
                reader = csv.reader(csvfile)
                for row in reader:
                    if row[0] == "P":
                        col_lidvids.add(row[1])

            # At this point we have a list of product lidvids that we need to find
            # in this directory or elsewhere.
            for p in src_col_dir.rglob("*.xml"):
                f = get_lidvidfile(p)
                f_lidvid = f["lid"] + "::" + f["vid"]
                if f_lidvid in col_lidvids:
                    dest_path = bld_col_dir / p.relative_to(src_col_dir)
                    dest_path.parent.mkdir(parents=True, exist_ok=True)
                    copy2(p, dest_path)
                    copy2(p.with_name(f["productfile"]), dest_path.parent)
                    col_lidvids.remove(f_lidvid)
                if len(col_lidvids) == 0:
                    break

            if len(col_lidvids) != 0:
                raise ValueError(f"Could not find the following lidvids: {col_lidvids}")

    return



#!/usr/bin/env python
# coding: utf-8

"""Defines the mapping or join table connecting the image_records and
pano_records tables using the SQLAlchemy ORM."""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from sqlalchemy import (
    ForeignKey,
    Integer,
)
from sqlalchemy.orm import mapped_column, relationship

from vipersci.vis.db import Base
from vipersci.vis.db.image_records import ImageRecord
from vipersci.vis.db.pano_records import PanoRecord


class JuncImagePano(Base):
    """An object to represent the junction of VIS Image Records to VIS Pano Records."""

    __tablename__ = "junc_image_pano"

    image_record_id = mapped_column(
        Integer, ForeignKey(ImageRecord.id), nullable=False, primary_key=True
    )
    pano_record_id = mapped_column(
        Integer, ForeignKey(PanoRecord.id), nullable=False, primary_key=True
    )

    image_record = relationship(
        "ImageRecord", back_populates="pano_record_associations"
    )
    pano_record = relationship("PanoRecord", back_populates="image_record_associations")



#!/usr/bin/env python
"""This module has tests for the nirvss_simulator module."""

# Copyright 2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.
import argparse
from pathlib import Path
import unittest
from unittest.mock import patch

import numpy as np

from vipersci.carto import nirvss_simulator as ns


class TestParser(unittest.TestCase):
    def test_parser(self):
        parser = ns.arg_parser()
        self.assertIsInstance(parser, argparse.ArgumentParser)


class TestLocationSimulator(unittest.TestCase):
    @patch("vipersci.carto.nirvss_simulator.rasterio.open")
    def test__init_map(self, mock_rasterio_data):
        ns.LocationSimulator._init_map(Path("dummy/path.raster"))
        mock_rasterio_data.transform.called_once()
        mock_rasterio_data.read.called_once_with(1)

    @patch(
        "vipersci.carto.nirvss_simulator.rasterio.transform.rowcol", return_value=(0, 0)
    )
    @patch(
        "vipersci.carto.nirvss_simulator.LocationSimulator._init_map",
        return_value=("xform", "array"),
    )
    def test_call(self, mock_init_map, mock_rowcol):
        simulator = ns.LocationSimulator(
            Path("dummy/bd.tif"), Path("dummy/insl.tif"), Path("dummy/temp.tif")
        )
        simulator.bd_arr = np.array([[2.5, 2.5], [2.5, 2.5]])
        simulator.insl_arr = np.array([[248, 248], [248, 248]])
        simulator.temp_arr = np.array([[264, 264], [264, 264]])

        h2o, oh = simulator(np.array([0, 0]))
        self.assertAlmostEqual(h2o, 4.371779725275095e-05)
        self.assertAlmostEqual(oh, 0.009899494936611665)



"""Creates VIS TIFF files from source 16-bit images.

This program is meant to help manually build test data sets.
"""

# Copyright 2022-2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
import logging
from pathlib import Path

from skimage.io import imread

from vipersci.pds import pid as pds
from vipersci.vis.create_image import write_tiff
from vipersci import util

logger = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__, parents=[util.parent_parser()]
    )
    parser.add_argument(
        "-p",
        "--product_id",
        required=True,
        help="The desired Product ID of the output image.",
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        type=Path,
        default=Path.cwd(),
        help="Output directory for TIFF file.",
    )
    parser.add_argument(
        "image", type=Path, help="Input image file, must be an unsigned 16-bit image."
    )
    return parser


def main():
    args = arg_parser().parse_args()
    util.set_logger(args.verbose)

    pid = pds.VISID(args.product_id)

    image = imread(str(args.image))

    write_tiff(pid, image, args.output_dir)

    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Determines accrual statistics based on geospatial vector files containing
regions of interest and a path geometry.

The geospatial files that can be provided are any that can be opened by
geopandas.read_file() which includes GeoPackages, Shapefiles, GeoJSON, and
more.
"""

# Copyright 2021-2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
from collections import Counter
from typing import Optional

import geopandas as gp
from shapely.geometry import LineString, box


def arg_parser():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "-a",
        "--areas",
        help="The geospatial file containing polygons or multi-polygons to "
        "evaluate the path against.  There should be one (multi)polygon "
        "for each category.",
    )
    parser.add_argument(
        "--iloc",
        default=0,
        type=int,
        help="If the path file has more than one geometry, select this "
        "indexed geometry.",
    )
    parser.add_argument("path", help="A geospatial file that contains geometries.")
    return parser


def main():
    parser = arg_parser()
    args = parser.parse_args()

    df = gp.read_file(args.path)
    path = df.iloc[args.iloc]
    # print(path)

    areas = gp.read_file(args.areas)

    if df.crs != areas.crs:
        parser.error(
            f"""\
The area and path have different coordinate reference systems.
    area CRS: {areas.crs}
    path CRS: {df.crs}
"""
        )

    try:
        accrual = accumulate(path["geometry"], areas)
    except ValueError as err:
        parser.error(err.msg)

    accumulated_length = sum(accrual.values())

    if accumulated_length != path["geometry"].length:
        print(f"Sum of the counted lengths: {accumulated_length}")
        print("This is different from the input total length!")

    print(f'Total length: {path["geometry"].length}')
    print(accrual)


def accumulate(
    path: LineString, areas: gp.GeoDataFrame, counter: Optional[dict] = None
) -> Counter:
    """Returns a collections.Counter of floating point accumulations
    of the length of *Path* against each of the geometries in *areas*.

    The *path* LineString is expected to be in the projection/CRS of *areas*.

    If a *counter* dict-like (which can be a collections.Counter object)
    is provided, returned accumulations will be added to those values
    provided in *counter*.
    """
    if not box(*areas.total_bounds).contains(path["geometry"]):
        raise ValueError(
            "The specified path geometry is not entirely contained within "
            "the area's bounding box."
        )

    if counter is None:
        counter = Counter()
    else:
        counter = Counter(counter)

    for area in areas.itertuples():
        counter[area.category] += path.intersection(area.geometry).length

    return counter



"""Contains information and functions that reflect information in the MCSE ICD.

Contains information about how to translate bit fields to their practical
values, as well as helper functions.
"""

# Copyright 2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import logging
from typing import Union


logger = logging.getLogger(__name__)


pga_gain_dict = {
    # The key is the integer value of the 3-bit parameter from the Image
    # Header.  The value is the floating point value that corresponds
    # to the 3-bit parameter.
    0: 1.0,
    1: 1.2,
    2: 1.4,
    3: 1.6,
    4: 2.0,  # 4, 5, 6, & 7 are 2x 0, 1, 2, 3
    5: 2.4,
    6: 2.8,
    7: 3.2,
}


def exposure_time(value: int):
    """Returns the exposure time in microseconds.

    The Exposure parameter is a 15 bit value, and the Exposure Step
    is a 1 bit value which is packed together so that the Exposure
    Step is the MSB of the two-byte value, defined as follows:

    The 15-bit LSB portion is "the exposure field" and the 1-bit Exposure Step
    modifies how the exposure field is interpreted.  If the Exposure Step is
    zero (0b0), the steps between exposure field values are 10 microseconds
    allowing values from 111 microseconds through ~328 microseconds.  If the
    Exposure Step is 1 (0b1), the steps between exposure field values are
    1 milisecond, from 1 milisecond to 32,768 ms.
    """

    if value <= 32768:
        return 111 + value * 10

    elif value < 65536:
        return (value - 32767) * 1000

    else:
        raise ValueError


def mcam_id(value: int):
    """Returns the camera value given the MCAM_ID.

    This is a 3-bit value  in the header.
    """

    d = {
        0: "Camera 1",
        1: "Camera 2",
        2: "Camera 3",
        3: "Camera 4",
        4: "Test image",
    }

    try:
        return d[value]
    except KeyError:
        raise ValueError(f"The value ({value}) must be in the set {d.keys()}.")


def pga_gain(value: Union[int, float]):
    """Returns the PGA Gain as a decimal value."""

    try:
        if isinstance(value, float):
            if value in pga_gain_dict.values():
                return value
            else:
                raise KeyError

        return pga_gain_dict[value]
    except KeyError:
        raise ValueError(
            f"The value ({value}) must be in the set {pga_gain_dict.keys()}."
        )



"""Creates a PDS4 Collection Inventory CSV file from the provided XML labels.
"""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import logging
from pathlib import Path

from vipersci.pds.labelmaker import get_lidvidfile, write_inventory
from vipersci import util

logger = logging.getLogger(__name__)


def add_parser(subparsers):
    parser = subparsers.add_parser("inventory", help=__doc__)

    parser.add_argument(
        "-n",
        "--name",
        required=True,
        help="The name that will be used to form the name of the output file.  If "
        "'-n foobar' is given, the output file will be collection_foobar.csv ",
    )
    parser.add_argument(
        "-m",
        "--member",
        choices=["P", "S"],
        default="P",
        help="Indicates whether the products are a primary (P) or secondary (S) member "
        "of the collection.  All products will be marked identically. "
        "Default: %(default)s",
    )
    parser.add_argument(
        "labelfiles",
        type=Path,
        nargs="+",
        help="Path(s) to all XML label files to be included in the output collection "
        "file.",
    )
    parser.set_defaults(func=main)
    return parser


def main(args):
    util.set_logger(args.verbose)

    outpath = Path(f"collection_{args.name}.csv")

    labels = []
    for path in args.labelfiles:
        labels.append(get_lidvidfile(path))

    write_inventory(outpath, labels, args.member)

    return



#!/usr/bin/env python
# coding: utf-8

"""Defines the VIS image_requests table using the SQLAlchemy ORM."""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Copyright (c) 2023, Million Concepts
# All rights reserved.
#
# This code is licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.
#
#
# The ImageRequest class below is a derived work based on the ImageRequest class from
# https://github.com/MillionConcepts/viper-vis-orchestrator/blob/main/viper_orchestrator/visintent/tracking/tables.py
# as of commit 897585605a7ace3b2f0886f1fd02776b3cd7245f (2023-09-24)
# written by Michael St. Clair.  That code is licensed under the BSD 3-Clause
# License:
#
# BSD 3-Clause License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from datetime import datetime
import enum
from typing import Sequence, Union

from sqlalchemy import (
    Boolean,
    DateTime,
    Enum,
    Identity,
    Integer,
    String,
)
from sqlalchemy.orm import mapped_column, relationship, validates
from geoalchemy2 import Geometry  # type: ignore

from vipersci.pds.datetime import isozformat
from vipersci.pds.pid import vis_instruments
from vipersci.vis.db import Base
from vipersci.vis.db.light_records import luminaire_names


class Status(enum.Enum):
    """
    This describes the status of an Image Request.
    """

    WORKING = 1
    READY_FOR_VIS = 2
    READY_FOR_PLANNING = 3
    PLANNED = 4
    NOT_PLANNED = 5
    IMMEDIATE = 6
    ACQUIRED = 7
    NOT_AQUIRED = 8
    NOT_OBTAINABLE = 9


class CameraType(enum.Enum):
    NAVCAM = 1
    AFTCAM = 2
    HAZCAM = 3


class CompressionType(enum.Enum):
    LOSSLESS = 1
    LOSSY = 2


class ImageMode(enum.Enum):
    # These are just for the NavCams and AftCams.
    LEFT = 1
    RIGHT = 2
    STEREO = 3
    PANORAMA = 4  # Only NavCams
    CALIBRATION = 5  # Only NavCams: initiates "standard" calibration sequence.


class RoverWaitFor(enum.Enum):
    DOWNLINK = 1
    VIS_VERIFICATION = 2
    DOWNLINK_AND_VIS = 3


class ImageRequest(Base):
    """An object to represent rows in the image_requests table for VIS."""

    __tablename__ = "image_requests"

    id = mapped_column(Integer, Identity(start=1), primary_key=True)
    title = mapped_column(String, nullable=False, doc="Short title for request.")
    justification = mapped_column(
        String, nullable=False, doc="Full description of request intent."
    )

    # TODO: determine whether to dynamically or statically update
    status = mapped_column(
        Enum(Status), nullable=False, default="WORKING", doc="request status"
    )
    users = mapped_column(String, nullable=False, doc="requesting user(s)")
    # note: autofilled; this will also track edit time / backfill time.
    request_time = mapped_column(
        DateTime(timezone=True), nullable=False, doc="time of request submission/update"
    )
    # all remaining fields may be null for backfilled requests. they are
    # intended for use in ops.
    target_location_point = mapped_column(
        Geometry(geometry_type="POINT", srid=910101),
        doc="If provided by the user the specific location of the image target point.",
    )
    target_location = mapped_column(
        String,
        doc="One-line description of what this should be an image "
        "of. May simply be coordinates from MMGIS.",
    )
    rover_location_point = mapped_column(
        Geometry(geometry_type="POINT", srid=910101),
        doc="If provided by the user the specific location of the rover for this "
        "image.",
    )
    rover_location = mapped_column(
        String,
        doc="One-line description of where the rover should be when "
        "the image is taken. May simply be coordinates from MMGIS.",
    )
    rover_orientation = mapped_column(
        String,
        default="any",
        doc="One-line description of desired rover orientation",
    )
    rover_wait_for = mapped_column(
        Enum(RoverWaitFor), doc="Details of a rover stop request."
    )
    # TODO: how do we handle attached images? are we responsible for file
    #  management?

    camera_type = mapped_column(Enum(CameraType), doc="Primary camera request.")
    imaging_mode = mapped_column(Enum(ImageMode), doc="Imaging mode to be used.")
    hazcams = mapped_column(
        String,
        doc="If camera_type is HAZCAM, this field should contain one or more hazcams.",
    )

    compression = mapped_column(
        Enum(CompressionType), doc="desired compression for image"
    )
    luminaires = mapped_column(
        String, default="default", doc="requested active luminaires"
    )
    exposure_time = mapped_column(String, default="default")

    # Start panorama-only parameters:
    caltarget_required = mapped_column(
        Boolean, default=True, doc="acquire caltarget image?"
    )
    aftcam_pair = mapped_column(
        Boolean, default=False, doc="acquire additional aftcam pair?"
    )
    chin_down_navcam_pair = mapped_column(
        Boolean, default=False, doc="acquire chin-down navcam pair?"
    )
    slices = mapped_column(Integer, doc="The total number of slices.")
    # needed only for non-360 panos, if blank, assume first_slice is "1" and
    # last slice is the number of slices, a.k.a a 360 pano.
    first_slice_index = mapped_column(Integer, doc="First slice of pano, >= 1.")
    last_slice_index = mapped_column(Integer, doc="Last slice of pano, <= slices.")
    # End panorama-only parameters.

    # These establish the many-to-many relationship between ImageRequests and LDST
    # hypotheses.  The junction table allows the relation to be marked "critical"
    # or not.
    ldst_hypotheses = relationship(
        "LDST",
        secondary="junc_image_request_ldst",
        back_populates="image_requests",
        viewonly=True,
    )
    ldst_associations = relationship(
        "JuncImageRequestLDST", back_populates="image_request"
    )

    # This establishes a many ImageRequests to one ImageRecord relationship.
    image_records = relationship("ImageRecord", back_populates="image_request")

    # legal values for things.
    # could do some of these as enum columns but very annoying for things that
    # start with numbers, generally more complicated, etc.

    # TODO: do we want a non-pano image request to be able to handle a series?

    luminaire_generalities = ("default", "none")

    @validates("luminaires")
    def validate_luminaires(self, _, value: Union[Sequence[str], str]):
        return self.validate_listing(
            "luminaires",
            ("default", "none"),
            tuple(luminaire_names.keys()),
            value,
            limit=4,
        )

    @validates("hazcams")
    def validate_hazcams(self, _, value: Union[Sequence[str], str]):
        return self.validate_listing(
            "hazcams",
            ("Any",),
            tuple(filter(lambda x: x.startswith("h"), vis_instruments.keys())),
            value,
        )

    @staticmethod
    def validate_listing(
        name: str,
        generalities: tuple,
        allowable: tuple,
        value: Union[Sequence[str], str],
        limit=None,
    ):
        if len(value) == 0:
            raise ValueError(
                f"At least '{generalities[0]}' must be specified for {name}."
            )

        if isinstance(value, str):
            value = value.split(",")

        if limit is not None and len(value) > limit:
            raise ValueError(f"Maximum {limit} {name}s per request.")
        elif len(value) == 1:
            legal = allowable + generalities
            if value[0] not in legal:
                raise ValueError(f"Single {name} entry must be one of {legal}.")
        else:
            if not all(v in allowable for v in value):
                raise ValueError(
                    f"Multi-element requests must be selected from:" f"{allowable}"
                )

        return ",".join(value)

    def asdict(self):
        d = {}

        for c in self.__table__.columns:
            if isinstance(getattr(self, c.name), datetime):
                d[c.name] = isozformat(getattr(self, c.name))
            else:
                d[c.name] = getattr(self, c.name)

        if hasattr(self, "labelmeta"):
            d.update(self.labelmeta)

        return d



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the `pds/xml` module."""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from textwrap import dedent
import unittest
import xml.etree.ElementTree as ET

from vipersci.pds import xml


class TestXML(unittest.TestCase):
    def setUp(self):
        self.xmltext = dedent(
            """\
        <?xml version="1.0" encoding="UTF-8"?>
        <?xml-model href="https://pds.nasa.gov/pds4/pds/v1/PDS4_PDS_1K00.sch"
          schematypens="http://purl.oclc.org/dsdl/schematron"?>
        <Product_Collection
          xmlns="http://pds.nasa.gov/pds4/pds/v1"
          xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
          xsi:schemaLocation="http://pds.nasa.gov/pds4/pds/v1
          https://pds.nasa.gov/pds4/pds/v1/PDS4_PDS_1K00.xsd">
        <File_Area_Inventory>
          <File>
            <file_name></file_name>
            <creation_date_time>2023-11-02T23:12:59.083415Z</creation_date_time>
          </File>
          <Inventory>
            <offset unit="byte">0</offset>
            <parsing_standard_id>PDS DSV 1</parsing_standard_id>
            <records>4</records>
            <record_delimiter>Carriage-Return Line-Feed</record_delimiter>
            <field_delimiter>Comma</field_delimiter>
          </Inventory>
        </File_Area_Inventory>
        </Product_Collection>
        """
        )

    def test_find_text(self):
        root = ET.fromstring(self.xmltext)
        self.assertRaises(
            ValueError,
            xml.find_text,
            root,
            ".//pds:File_Area_Inventory/pds:Inventory/pds:offset",
            unit_check="bit",
        )

        self.assertRaises(
            ValueError,
            xml.find_text,
            root,
            ".//pds:File_Area_Inventory/pds:File/pds:file_name",
        )

        self.assertRaises(
            ValueError,
            xml.find_text,
            root,
            ".//pds:File_Area_Inventory/pds:File/pds:md5",
        )



#!/usr/bin/env python
"""This module contains viss utility functions."""

# Copyright 2021-2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
import hashlib
import logging
import os

import vipersci


def md5(path: os.PathLike):
    """Returns a string objects of hexadecimal digits representing the md5 hash
    of the file at *path*."""
    m = hashlib.md5()
    with open(path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            m.update(chunk)

    return m.hexdigest()


def parent_parser() -> argparse.ArgumentParser:
    """Returns a parent parser with common arguments for viss programs."""
    parent = argparse.ArgumentParser(add_help=False)
    parent.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Displays additional information.",
    )
    parent.add_argument(
        "--version",
        action="version",
        version=f"vipersci Software version {vipersci.__version__}",
        help="Show library version number.",
    )
    return parent


def set_logger(verblvl=None) -> None:
    """Sets the log level and configuration for applications."""
    logger = logging.getLogger(__name__.split(".")[0])
    lvl_dict = {0: logging.WARNING, 1: logging.INFO, 2: logging.DEBUG}
    if verblvl in lvl_dict:
        lvl = lvl_dict[verblvl]
    else:
        lvl = lvl_dict[max(lvl_dict.keys())]

    logger.setLevel(lvl)

    ch = logging.StreamHandler()
    ch.setLevel(lvl)

    if lvl < 20:  # less than INFO
        formatter = logging.Formatter("%(name)s - %(levelname)s: %(message)s")
    else:
        formatter = logging.Formatter("%(levelname)s: %(message)s")

    ch.setFormatter(formatter)
    logger.addHandler(ch)

    return



"""Inspects VIS and other images.
"""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
import json
import logging
from typing import Union
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from scipy import stats
from skimage.exposure import equalize_adapthist
from skimage.io import imread

from vipersci.vis import image_statistics
from vipersci import util

logger = logging.getLogger(__name__)

ImageType = Union[npt.NDArray[np.uint16], npt.NDArray[np.uint8]]


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__, parents=[util.parent_parser()]
    )
    parser.add_argument(
        "--clahe",
        action="store_true",
        help="Apply Contrast Limited Adaptive Histogram Equalization (CLAHE) to the "
        "input image.",
    )
    parser.add_argument(
        "-g",
        "--grid",
        action="store_true",
        help="If given will display the OpenMCT 'grid' on the image.",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        default=Path.cwd(),
        help="If given, will write output image to this file.",
    )
    parser.add_argument("--vmin", type=float, help="Minimum DN value to display.")
    parser.add_argument("--vmax", type=float, help="Maximum DN value to display.")
    parser.add_argument("input", type=Path, help="File containing image or JSON label.")
    return parser


def main():
    args = arg_parser().parse_args()
    util.set_logger(args.verbose)

    if ".json" == args.input.suffix.lower():
        with open(args.input, "r") as f:
            info = json.load(f)

            p = Path(info["file_path"])
            if not p.is_absolute():
                p = args.input.parent / p

        logger.info(f"Reading {p}")
        image = imread(p)
        imtitle = info["product_id"]
    else:
        image = imread(args.input)
        imtitle = args.input.name

    logger.info(describe(image, "image as loaded:"))
    logger.info(image_statistics.pprint(image))

    if args.clahe:
        image = equalize_adapthist(image)
        logger.info(describe(image, "image after CLAHE:"))

    plot_img_and_hist(
        image, title=imtitle, vmin=args.vmin, vmax=args.vmax, grid=args.grid, save=False
    )

    return


def describe(image: ImageType, message: str):
    lines = list()
    d = stats.describe(image.ravel())
    lines.append(message)
    lines.append(f"  shape: {image.shape}  dtype: {image.dtype}")
    lines.append(f"  minmax: {d.minmax}  mean: {d.mean}  variance: {d.variance}")
    # lines.append("  " + pprint.pformat(
    #     stats.describe(image.ravel()), indent=2
    # ))
    return "\n".join(lines)


def plot_img_and_hist(
    image, images=None, title="Image", vmin=None, vmax=None, grid=False, save=False
):
    """Plot an image along with its histogram"""
    if images is None:
        images = [
            image,
        ]
    else:
        images.insert(0, image)

    fig, axes = plt.subplots(1, len(images) + 1)

    ax_images = axes[:-1]
    ax_hist = axes[-1]

    fig.suptitle(title)

    # Display image
    for ax, im in zip(ax_images, images):
        ax.imshow(
            im,
            cmap=plt.cm.gray,
            vmin=vmin if vmin is not None else np.amin(im),
            vmax=vmax if vmax is not None else np.amax(im),
        )
        if grid:
            ax.axhline(im.shape[0] / 4, color="blue")
            ax.axhline(im.shape[0] / 2, color="blue")
            ax.axhline(im.shape[0] * 3 / 4, color="blue")
            ax.axvline(im.shape[1] / 4, color="blue")
            ax.axvline(im.shape[1] / 2, color="blue")
            ax.axvline(im.shape[1] * 3 / 4, color="blue")
        # ax.set_axis_off()

    # Display histogram
    # ax_hist.hist(image.ravel(), bins=bins, histtype="bar", log=True)
    hist_dn, hist_counts = np.unique(image, return_counts=True)
    ax_hist.fill_between(hist_dn, hist_counts, alpha=0.2, color="C0")
    ax_hist.scatter(hist_dn, hist_counts, marker=".")
    ax_hist.set_yscale("log")
    ax_hist.set_xlabel("DN")
    ax_hist.set_ylabel("Pixel Count")
    if vmin is not None:
        ax_hist.axvline(vmin, color="gray")
    if vmax is not None:
        ax_hist.axvline(vmax, color="gray")

    fig.tight_layout()
    # fig.set_dpi(300)
    # fig.set_size_inches(w=6.4, h=4.8)
    plt.show()

    if save:
        fname = input("Filename (blank to not save):")
        if len(fname) > 0:
            fig.savefig(fname, dpi=300)



#!/usr/bin/env python
"""This module has tests for the vis.pds.labelmaker.inventory functions."""

# Copyright 2023, vipersci developers.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import unittest
from unittest.mock import patch

import vipersci.pds.labelmaker.cli as cli


class TestMain(unittest.TestCase):
    @patch("vipersci.pds.labelmaker.cli.argparse.ArgumentParser", autospec=True)
    def test_main(self, m_parser):
        cli.main()

        m_parser.assert_called()



"""Check to see if images could be made into a panorama.
"""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
from datetime import datetime, timezone
from functools import partial
import itertools
import logging
from pathlib import Path
from typing import Iterable, Union

import pandas as pd
import requests
from sqlalchemy import create_engine, select
from sqlalchemy.orm import Session

from vipersci.vis.db.image_records import ImageRecord
from vipersci.pds import pid as pds
from vipersci import util

logger = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__, parents=[util.parent_parser()]
    )
    parser.add_argument(
        "--csv",
        type=Path,
        help="Path to CSV file with times and locations.  Ignored if --url is set.",
    )
    parser.add_argument(
        "-d",
        "--dburl",
        help="Database with an image_records table to read from. "
        "If not given, no database will be written to.  Example: "
        "postgresql://postgres:NotTheDefault@localhost/visdb",
    )
    parser.add_argument(
        "-s",
        "--since",
        default=0,
        help="An ISO8601 datetime after which should be searched for images "
        "in the database.",
    )
    parser.add_argument(
        "-u",
        "--until",
        default=datetime.now(tz=timezone.utc),
        help="An ISO8601 datetime before which should be searched for images "
        "in the database.",
    )
    parser.add_argument(
        "--url",
        help="URL to which event_times may be posted and for which location and "
        "orientation will be returned.",
    )
    parser.add_argument(
        "--input", nargs="*", required=False, help="VIS Image Record Product IDs."
    )
    return parser


def main():
    parser = arg_parser()
    args = parser.parse_args()
    util.set_logger(args.verbose)

    gppf = None
    if args.url is not None:
        gppf = partial(get_position_and_pose_from_mapserver, url=args.url)
    elif args.csv is not None:
        gppf = partial(get_position_and_pose_from_csv, path=args.csv)
    else:
        parser.error(
            "Neither --url nor --csv were given.  Need at least one to be a "
            "source for position and pose with time."
        )

    if args.input is not None:
        pids = list(map(pds.VISID, args.input))
        pano_groups = check(pids, gppf)

    else:
        if args.dburl is None:
            parser.error("To query images between times, you must enter a --dburl.")
        t_since = datetime.fromisoformat(args.since)
        t_until = datetime.fromisoformat(args.until)
        if args.until < args.since:
            parser.error(
                f"The -s argument ({args.since}) must be less than the -u "
                f"argument ({args.until})."
            )

        engine = create_engine(args.dburl)
        with Session(engine) as session:
            result = session.scalars(
                select(ImageRecord).where(
                    t_since <= ImageRecord.start_time, ImageRecord.start_time <= t_until
                )
            )
            pano_groups = check(result, gppf)

    for pg in pano_groups:
        print(f"For position and pose {pg[1]}:")
        for p in pg[0]:
            print(f"\t{p}")

    return


def check(
    images: Iterable[Union[pds.VISID, ImageRecord]], get_pos_pose_func=None
) -> list:
    """
    Returns a list of two-tuples.  The first element in each two-tuple is a list of
    items from *images* which could be made into a panorama, and the second element
    is the position and pose associated with those *images*.

    The *get_pos_pose_func* should be a function that takes a list of timestamp times
    and returns an object that represents a position and pose at that time (please see
    get_position_and_pose_from_df() and get_position_and_pose_from_mapserver() in this
    module.  These functions may need to be wrapped via functools.partial() so that the
    function passed here takes only a list of timestamp times.

    If the iterable does not contain all ImageRecord objects or all VISID objects,
    a TypeError will be raised.
    """
    if get_pos_pose_func is None:
        raise ValueError("A function must be provided to get_pos_pose_func.")

    raw_vids = []
    image_records = None
    if all(map(isinstance, images, itertools.repeat(ImageRecord))):
        image_records = {}
        for im in images:
            vid = pds.VISID(im.product_id)  # type: ignore  # noqa
            image_records[vid] = im
            raw_vids.append(vid)
    elif all(map(isinstance, images, itertools.repeat(pds.VISID))):
        raw_vids = list(images)  # type: ignore  # noqa
    else:
        raise TypeError(
            "The provided iterable does not contain all ImageRecords or "
            f"all VISIDs, it is {images}."
        )

    raw_vids.sort()
    vids = pds.VISID.best_compression(raw_vids)

    vids_by_time = {}
    times = []
    for v in vids:
        if v.compression == "s":
            continue
        ts = v.datetime().timestamp()
        if ts not in vids_by_time:
            vids_by_time[ts] = v
            times.append(ts)

    tpp = get_pos_pose_func(times)
    grouped = groupby_2nd(tpp)

    pano_groups = list()
    for t_list, position_pose in grouped:
        if len(t_list) > 1:
            # We have a pano group
            if image_records is not None:
                ids = [image_records[vids_by_time[t]] for t in t_list]
            else:
                ids = [vids_by_time[t] for t in t_list]
            pano_groups.append((ids, position_pose))

    return pano_groups


def get_position_and_pose_from_csv(
    times: list,
    path: Path,
    time_column=0,
    x_column=1,
    y_column=2,
    yaw_column=3,
):
    """
    Given a list of timestamp times and a Path to a CSV file with the specified columns,
    return a list of two-tuples whose first element is the time and whose
    second element is three-tuple of x-location, y-location, and yaw.

    If None is given for any of the x-, y-, or yaw- columns, they are ignored
    from the CSV read.
    """
    uc = list()
    for c in (time_column, x_column, y_column, yaw_column):
        if c is not None:
            uc.append(c)

    df = pd.read_csv(path, usecols=uc)

    tpp = list()
    for t in times:
        row = df.iloc[(df.iloc[:, time_column] - t).abs().argsort()[:1]]
        tpp.append((t, tuple(row.values.flatten().tolist()[1:])))

    return tpp


def get_position_and_pose_from_mapserver(
    times: list, url: str, crs: str = "VIPER:910101"
):
    """
    Given a list of times and a URL that requests can be made against,
    return a list of two-tuples whose first element is the time and
    whose second element is a three-tuple of x-location, y-location,
    and yaw.
    """

    tpp = list()

    for t in times:
        position_result = requests.get(
            url,
            params={
                "event_time": t,
                "crs_code": crs,
            },
        )
        rj = position_result.json()

        tpp.append((t, (rj["location"][0], rj["location"][1], rj["yaw"])))

    return tpp


def groupby_2nd(
    tuples: list,
):
    """
    Given a list of sorted two-tuples, determine groupings based on the
    second element of the two-tuple.

    The returned list of two-tuples contains one "group" per element.  The
    first element of the two tuple is the list of input first elements that have
    the same input second element, and the second element of the two-tuple is the
    input second element.

    The list should be sorted by increasing order on the second element of the tuple.

    For example::

        >>> t = [("Alice", "foo"), ("Bob", "bar"), ("Catherine", "bar")]
        >>> g = groupby_2nd(t)
        >>> print(g)
        [(["Alice",], "foo"), (["Bob", "Catherine"], "bar")]
    """

    def keyfunc(t: tuple):
        return t[1]

    grouped = list()
    for k, g in itertools.groupby(tuples, key=keyfunc):
        first = list()
        second = ""
        for elem in g:
            first.append(elem[0])
            second = elem[1]
        grouped.append((first, second))

    return grouped



#!/usr/bin/env python
"""This module contains an alternate form of the vipersci.pds.datetime.isozformat()
function suitable for testing with SQLite databases which do not retain timezone
information.
"""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import datetime


def isozformat(date_time, sep="T", timespec="auto"):
    """
    This function is meant to be patched in place of vipersci.pds.datetime.isozformat()
    via unittest.mock.patch's 'new=' argument as follows:

        with patch(
            "vipersci.vis.db.image_records.isozformat",
            new=datetime_sqlite.isozformat
        )

    It should only be patched in when needed to test load-from-database functionality.
    If used more frequently, could allow bugs to propagate.

    For more information about isozformat, please see vipersci.pds.datetime.isozformat()
    """
    if date_time.utcoffset() == datetime.timedelta():
        return date_time.replace(tzinfo=None).isoformat(sep, timespec) + "Z"
    elif date_time.utcoffset() is None:
        # This elif is the difference from the original function: it will take a naive
        # datetime and assumes that it is a UTC datetime
        return date_time.isoformat(sep, timespec) + "Z"
    else:
        raise ValueError(
            "The datetime object is either naive (not timezone aware), "
            "or has a non-zero offset from UTC.  Maybe you just want "
            "the datetime object's isoformat() function?"
        )



"""Check to see if an image has anomalous pixels which may indicate that those pixels
in the detector may be of concern.
"""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
import logging
from pathlib import Path
import sys

import numpy as np
from skimage.filters import median
from skimage.io import imread

from vipersci import util

logger = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__, parents=[util.parent_parser()]
    )
    parser.add_argument(
        "-t",
        "--tolerance",
        type=float,
        default=3,
        help="The number of standard deviations of the difference between the image "
        "and its median filter to 'trigger' on.",
    )
    parser.add_argument("input", type=Path, help="VIS Image.")
    return parser


def main():
    parser = arg_parser()
    args = parser.parse_args()
    util.set_logger(args.verbose)

    image = imread(args.input)

    indices = check(image, args.tolerance)

    print(f"There are {len(indices[0])} anomalous pixels.")
    print(f"or {100 * len(indices[0]) / image.size} %")
    print(indices)

    return


def check(image, tolerance=3):
    """
    Returns indices of the elements of *image* which exceed *tolerance* standard
    devations of the difference between *image* and a median-filtered version of it.
    """

    blurred = median(image)
    difference = image.astype(int) - blurred.astype(int)
    threshold = tolerance * np.std(difference)
    logger.info(f"threshold: {threshold}")

    anom_pixel_indices = np.nonzero(np.abs(difference) > threshold)

    return anom_pixel_indices


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the `util` module."""

# Copyright 2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
import logging
import unittest
from unittest.mock import mock_open, patch

import vipersci.util as util


class TestUtil(unittest.TestCase):
    def test_md5(self):
        with patch("vipersci.util.open", mock_open(read_data=b"test")):
            h = util.md5("bogus.file")
            self.assertEqual("098f6bcd4621d373cade4e832627b4f6", h)

    def test_parent_parser(self):
        self.assertIsInstance(util.parent_parser(), argparse.ArgumentParser)

    def test_logging(self):
        util.set_logger(verblvl=0)
        logger = logging.getLogger()
        self.assertEqual(30, logger.getEffectiveLevel())



#!/usr/bin/env python
"""This module has tests for the vis.pds.labelmaker.inventory functions."""

# Copyright 2023, vipersci developers.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from argparse import ArgumentParser
import unittest
from unittest.mock import patch

import vipersci.pds.labelmaker.inventory as inv
from vipersci import util


class TestParser(unittest.TestCase):
    def test_add_parser(self):
        parser = ArgumentParser()
        subparsers = parser.add_subparsers()
        inv.add_parser(subparsers)

        d = vars(
            parser.parse_args(
                ["inventory", "--name", "dumbname", "file1.xml", "file2.xml"]
            )
        )
        self.assertIn("name", d)
        self.assertIn("member", d)
        self.assertIn("labelfiles", d)

    def test_main(self):
        parser = ArgumentParser(parents=[util.parent_parser()])
        subparsers = parser.add_subparsers()
        inv.add_parser(subparsers)

        args = parser.parse_args(
            ["inventory", "--name", "dumbname", "file1.xml", "file2.xml"]
        )

        with patch(
            "vipersci.pds.labelmaker.inventory.get_lidvidfile",
            return_value={"dummy": "dict"},
        ) as m_get_lidvid, patch(
            "vipersci.pds.labelmaker.inventory.write_inventory",
        ) as m_write_inv:
            inv.main(args)

            self.assertEqual(2, m_get_lidvid.call_count)
            m_write_inv.assert_called_once()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Displays color maps and can create output colorization files.
"""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
from io import StringIO
import logging
from pathlib import Path
from typing import Sequence, Union

import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np

from vipersci import __version__
from vipersci import util

logger = logging.getLogger(__name__)

presets = dict(
    dd=dict(
        label="Depth to Diameter ratio",
        cmap="viridis",
        vmin=0.05,
        vmax=0.15,
        bounded=False,
        extend="both",
    ),
    dice=dict(
        label="Depth to Ice Stability (m)",
        cmap="Blues_r",
        vmin=0,
        vmax=1,
        bounded=False,
        extend="neither",
    ),
    isr=dict(
        label="Depth to Ice Stability (m)",
        cmap="Blues_r",
        vmin=0,
        vmax=1,
        bounded=[0, 0.01, 0.2, 0.85, 1.0],
        extend="neither",
    ),
    stoplight_slope=dict(
        label="Slope (°)", cmap="RdYlGn_r", vmin=0, vmax=20, bounded=False, extend="max"
    ),
    slope=dict(
        label="Slope (°)", cmap="RdPu", vmin=0, vmax=20, bounded=False, extend="max"
    ),
    slope_disc=dict(
        label="Slope (°)",
        cmap="RdPu",
        vmin=0,
        vmax=20,
        bounded=[0, 3, 5, 10, 15, 20],
        extend="max",
    ),
    tmax=dict(
        label="Max Temperature (K)",
        cmap="inferno",
        vmin=22,
        vmax=351,
        bounded=False,
        extend="neither",
    ),
    verve_slope=dict(
        label="Slope (°)", cmap="", vmin=0, vmax=20, bounded=False, extend="max"
    ),
    verve_slope_discrete=dict(
        label="Slope (°)",
        cmap="",
        vmin=0,
        vmax=20,
        bounded=[0, 5, 10, 15, 20],
        extend="max",
    ),
    weh=dict(
        label="Water Equivalent Hydrogen fraction",
        cmap="BuGn",
        vmin=0,
        vmax=1,
        bounded=False,
        extend="neither",
    ),
)


class Palette:
    """
    This class maintains information about a specific instance of a color map for a
    particular purpose.
    """

    def __init__(
        self,
        cmap,
        vmin,
        vmax,
        label="Some Units",
        bounded: Union[bool, Sequence] = False,
        extend="neither",
        nodata_color="none",
        under_color=0,
        over_color=0.999999,
    ):
        self.nodata_color = mpl.colors.to_rgba(nodata_color)
        if isinstance(cmap, mpl.colors.Colormap):
            self.cmap = cmap
        else:
            self.cmap = mpl.colormaps[cmap]

        self.cmap.colorbar_extend = extend

        if isinstance(under_color, (float, int)):
            under_color = self.cmap(under_color)

        if isinstance(over_color, (float, int)):
            over_color = self.cmap(over_color)

        if extend == "max":
            self.cmap.set_under(under_color, alpha=0)
        elif extend == "neither":
            self.cmap.set_under(under_color, alpha=0)
            self.cmap.set_over(over_color, alpha=0)
        elif extend == "both":
            pass
        else:
            raise ValueError(f"extend={extend} is not accepted.")
        self.vmin = vmin
        self.vmax = vmax
        self.label = label
        self.bounded = bounded
        if bounded:
            self.norm = mpl.colors.BoundaryNorm(bounded, self.cmap.N, extend=extend)
        else:
            self.norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)

    def to_gdal_colormap(self, epsilon=0.001):
        s = [
            f"# Color table for {self.label}",
            f"# created by {__name__} version {__version__}",
            "# value, R, G, B, A",
        ]
        if self.bounded:
            if self.cmap.colorbar_extend in ("max", "neither"):
                values = [self.bounded[0] - epsilon]
            else:
                values = list()

            values.append(self.bounded[0])
            for v in self.bounded[1:-1]:
                values.append(v - epsilon)
                values.append(v)

            if self.cmap.colorbar_extend in ("min", "max", "neither"):
                values.append(self.bounded[-1] - epsilon)

            values.append(self.bounded[-1])
        else:
            values = np.linspace(self.vmin, self.vmax, num=self.cmap.N)
            if self.cmap.colorbar_extend in ("max", "neither"):
                values = np.insert(values, 0, self.vmin - epsilon)

            if self.cmap.colorbar_extend in ("min", "neither"):
                values = np.append(values, self.vmax + epsilon)

        for v in values:
            rgba = self.cmap(self.norm(v), bytes=True)
            s.append(f"{v} {' '.join(map(str, rgba))}")

        if not mpl.colors.same_color(self.nodata_color, (0, 0, 0, 0)):
            s.append(f"nv {' '.join(map(str, color_to_bytes(self.nodata_color)))}")

        return "\n".join(s)


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__, parents=[util.parent_parser()]
    )

    parser.add_argument(
        "-b",
        "--bar",
        choices=["v", "h"],
        help="If given will plot just a colorbar. Must specify h (horizontal) or v "
        "(vertical).",
    )
    parser.add_argument(
        "-c",
        "--colormap",
        help="Name of matplotlib colormap.",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        help="Optional name of output file.",
        required=False,
    )
    parser.add_argument(
        "-p",
        "--preset",
        choices=presets.keys(),
        required=False,
        help="Specifying a preset, sets the colormap, vmin, and vmax (ignoring those "
        "arguments if provided).",
    )
    parser.add_argument(
        "--plot",
        action="store_true",
        help="If given will create example plot, ignores -b.",
    )
    parser.add_argument(
        "--vmin", default=0, type=float, help="Minimum value for data range."
    )
    parser.add_argument(
        "--vmax", default=1, type=float, help="Maximum value for data range."
    )
    return parser


def main():
    parser = arg_parser()
    args = parser.parse_args()
    util.set_logger(args.verbose)

    if args.preset:
        if args.preset in ("verve_slope", "verve_slope_discrete"):
            d = presets[args.preset]
            d["cmap"] = verve_stoplight()
            pal = Palette(**d)
        else:
            pal = Palette(**presets[args.preset])
    else:
        pal = Palette(
            cmap=mpl.colormaps[args.colormap],
            vmin=args.vmin,
            vmax=args.vmax,
        )

    if args.plot:
        example_plot(pal, args.output)

    elif args.bar is not None:
        plot_colorbar(pal, args.bar, args.output)

    else:
        if args.output is None:
            print(pal.to_gdal_colormap())
        else:
            args.output.write_text(pal.to_gdal_colormap())


def color_to_bytes(color_tuple):
    """
    Returns tuple of values from 0 to 255, based on values in input tuple assumed
    to be in the range zero to one.
    """
    converted = map(lambda x: int(x * 255), color_tuple)
    return tuple(converted)


def example_plot(palette: Palette, output=None):
    """Apply colors to 2d function for example visualization."""

    plt.ioff()
    # data = rescale(moon(), pal.vmin, pal.vmax)  # (512, 512) uint8 ndarray
    # make these smaller to increase the resolution
    dx = 0.05
    x = np.arange(-3.0, 3.0, dx)
    x_grid, y_grid = np.meshgrid(x, x)
    z = rescale(
        (1 - x_grid / 2 + x_grid**5 + y_grid**3) * np.exp(-(x_grid**2 + y_grid**2)),
        palette.vmin,
        palette.vmax,
    )
    extent = np.min(x_grid), np.max(x_grid), np.min(y_grid), np.max(y_grid)

    checkerboard = np.add.outer(range(32), range(32)) % 2
    plt.imshow(
        checkerboard,
        cmap=plt.cm.gray,
        vmin=-1,
        vmax=2,
        interpolation="nearest",
        extent=extent,
    )

    # plt.imshow(data, cmap=pal.cmap, norm=pal.norm)
    plt.imshow(
        z, cmap=palette.cmap, norm=palette.norm, interpolation="bilinear", extent=extent
    )

    plt.colorbar(label=palette.label)
    plt.title(f"Colormap: {palette.cmap.name}")
    plt.tick_params(
        left=False, right=False, labelleft=False, labelbottom=False, bottom=False
    )
    plt.xlabel(
        f"Colormap range: {palette.vmin} to {palette.vmax}, "
        f"Data range: {z.min():.3g} to {z.max():.3g}"
    )

    if output is None:
        plt.show()
    else:
        plt.savefig(output)

    return


def plot_colorbar(palette: Palette, orientation: str = "h", output=None):
    """Plot colorbars only."""

    o_dict = {"h": "horizontal", "v": "vertical"}
    if orientation in o_dict:
        orientation = o_dict[orientation]

    plt.ioff()
    if orientation == "horizontal":
        fig, ax = plt.subplots(figsize=(5, 1))
        fig.subplots_adjust(bottom=0.5)
    elif orientation == "vertical":
        fig, ax = plt.subplots(figsize=(1.2, 5))
        fig.subplots_adjust(right=0.4)

    fig.colorbar(
        mpl.cm.ScalarMappable(norm=palette.norm, cmap=palette.cmap),
        cax=ax,
        orientation=orientation,
        label=palette.label,
    )

    if output is None:
        plt.show()
    else:
        plt.savefig(output)

    return


def rescale(arr, min, max, range_mult=0.1):
    """
    Returns numpy array of data based on array, but scaled to be range_mult smaller and
    larger than min and max,
    """
    data = arr.astype("float64")
    minmax_range = max - min
    real_min = min - (minmax_range * range_mult)
    real_max = max + (minmax_range * range_mult)
    return (
        ((data - np.min(data)) / (np.max(data) - np.min(data))) * (real_max - real_min)
    ) + real_min


def verve_stoplight():
    """Returns VERVE stoplight colors as a matplotlib colormap object."""
    color_arr = np.flipud(
        np.loadtxt(
            StringIO(
                """100 254 14 2
99 253 20 4
98 253 27 5
97 253 33 7
96 252 39 9
95 252 45 10
94 252 51 12
93 252 57 13
92 251 63 15
91 251 69 16
90 250 74 18
89 250 80 20
88 249 86 21
87 249 90 22
86 249 96 24
85 249 102 25
84 249 106 27
83 248 112 29
82 247 117 30
81 247 122 32
80 247 126 33
79 246 132 34
78 246 136 35
77 246 141 37
76 245 145 39
75 245 150 40
74 245 154 42
73 244 159 43
72 244 163 45
71 244 167 46
70 243 171 47
69 243 176 49
68 242 180 51
67 242 183 52
66 242 188 53
65 242 192 55
64 241 195 56
63 241 199 58
62 241 203 59
61 240 206 60
60 240 209 62
59 239 213 63
58 239 216 64
57 238 220 67
56 239 223 68
55 238 226 68
54 237 228 70
53 237 232 72
52 236 235 73
51 235 236 75
50 231 236 76
49 229 235 77
48 225 236 79
47 222 234 80
46 218 234 81
45 215 234 82
44 213 234 83
43 209 234 85
42 207 233 87
41 203 233 87
40 201 233 89
39 198 232 90
38 195 232 92
37 193 231 93
36 190 231 95
35 187 231 96
34 185 231 96
33 183 230 98
32 180 230 99
31 178 229 100
30 176 229 102
29 174 229 103
28 172 228 105
27 169 228 106
26 168 228 107
25 166 227 108
24 164 227 109
23 162 226 110
22 160 225 112
21 159 226 113
20 157 225 114
19 156 225 116
18 154 224 117
17 153 224 118
16 152 224 119
15 150 224 121
14 149 223 121
13 148 222 123
12 146 222 124
11 146 222 125
10 144 221 126
9 143 221 127
8 143 221 128
7 142 220 130
6 141 220 130
5 141 220 132
4 140 220 133
3 139 220 134
2 139 219 135
1 139 219 135
0 139 219 135"""
            ),
            dtype=np.dtype(int),
            usecols=(1, 2, 3),
        )
    )

    return mpl.colors.ListedColormap(color_arr / 255, name="VERVE_stoplight")



"""Creates Raw VIS PDS Products.

This module builds "raw" VIS data products from VIS record data.  At this time,
a Raw Product is produced from a VIS Image Record and zero to many Light Records.

A VIS Image Record consists of a TIFF file and metadata (either as a JSON file or
in a database).  That metadata is the basis for production of the PDS4 XML label file.

In order to perform the lookup of Light Records for an Image Record requires use
of the database.

For now, this program still has a variety of hard-coded elements,
that will eventually be extracted from telemetry.

The command-line version is primarily to aide testing.
"""

# Copyright 2022-2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
from datetime import datetime, timedelta, timezone
import json
import logging
from typing import Union
from pathlib import Path
from warnings import warn

import numpy as np
import numpy.typing as npt
from geoalchemy2 import load_spatialite  # type: ignore
from sqlalchemy import and_, create_engine, select
from sqlalchemy.event import listen
from sqlalchemy.orm import Session

import vipersci
from vipersci.vis.db.image_records import ImageRecord, ProcessingStage
from vipersci.vis.db.light_records import (
    LightRecord,
    luminaire_names,
    luminaire_shortnames,
)
from vipersci.vis.create_image import tif_info
from vipersci.vis.pds import lids, write_xml
from vipersci.pds import pid as pds
from vipersci.pds.datetime import isozformat
from vipersci import util

logger = logging.getLogger(__name__)

ImageType = Union[npt.NDArray[np.uint16], npt.NDArray[np.uint8]]


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__, parents=[util.parent_parser()]
    )
    parser.add_argument(
        "-d",
        "--dburl",
        required=True,
        help="Database with a raw_products table which will be written to. "
        "If not given, no database will be written to.  Example: "
        "postgresql://postgres:NotTheDefault@localhost/visdb",
    )
    parser.add_argument(
        "-t",
        "--template",
        default="raw-template.xml",
        help="Genshi XML file template.  Will default to the raw-template.xml "
        "file distributed with the module.  Only relevant when --xml is provided.",
    )
    parser.add_argument(
        "--tiff",
        type=Path,
        help="Optional pre-existing TIFF file (presumably created by create_image). "
        "This file will be inspected and its information added to the output. ",
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        type=Path,
        default=Path.cwd(),
        help="Output directory for label. Defaults to current working directory.  "
        "Output file names are fixed based on product_id, and will be over-written.",
    )
    parser.add_argument(
        "input",
        help="Product ID (or TIFF file from which a Product ID can be extracted) or a "
        "JSON file containing metadata.",
    )
    return parser


def main():
    parser = arg_parser()
    args = parser.parse_args()
    util.set_logger(args.verbose)

    engine = create_engine(args.dburl)
    if args.dburl.startswith("sqlite://"):
        listen(engine, "connect", load_spatialite)

    with Session(engine) as session:
        try:
            pid = pds.VISID(args.input)
        except ValueError:
            pid = None

        ir = None
        if pid is None or args.input.endswith(".json"):
            if Path(args.input).exists():
                with open(args.input) as f:
                    ir = ImageRecord(**json.load(f))
            else:
                parser.error(f"The file {args.input} does not exist.")
        else:
            # We got a valid pid, go look it up in the db.
            stmt = select(ImageRecord).where(ImageRecord.product_id == str(pid))
            result = session.scalars(stmt)
            rows = result.all()
            if len(rows) > 1:
                raise ValueError(f"There was more than 1 row returned from {stmt}")
            elif len(rows) == 0:
                raise ValueError(
                    f"No records were returned from the database for Product Id {pid}."
                )

            ir = rows[0]

        if ir is None:
            raise ValueError(f"Could not extract an ImageRecord from {args.input}")

        # I'm not sure where these are coming from, let's hard-code them for now:
        metadata = {
            "mission_phase": "TEST",
            "bad_pixel_table_id": 0,
        }

        # This allows values in these dicts to override the hard-coded values above.
        metadata.update(label_dict(ir, get_lights(ir)))
        if args.dburl.startswith("sqlite://"):
            for c in ir.__table__.columns:
                dt = getattr(ir, c.name)
                if isinstance(dt, datetime) and dt.tzinfo is None:
                    setattr(ir, c.name, dt.replace(tzinfo=timezone.utc))

        metadata.update(ir.asdict())
        metadata.update(
            {
                "software_name": "vipersci",
                "software_version": vipersci.__version__,
                "software_type": "Python",
                "software_program_name": __name__,
            }
        )
        if metadata["verification_purpose"] is None:
            metadata["purpose"] = "Science"
        else:
            metadata["purpose"] = (
                metadata["verification_purpose"].value.replace("_", " ").title()
            )

    if args.input.endswith(".tif"):
        args.tiff = Path(args.input)

    if args.tiff is not None:
        t_info = tif_info(args.tiff)
    else:
        t_info = tif_info(Path(ir.file_path))

    for k, v in t_info.items():
        if hasattr(metadata, k):
            if metadata[k] != v:
                raise ValueError(
                    f"The value of {k} in the metadata ({metadata[k]}) does not match "
                    f"the value ({v}) in the image ({args.tiff})"
                )
        else:
            if isinstance(v, datetime):
                metadata[k] = isozformat(v)
            else:
                metadata[k] = v

    write_xml(metadata, args.template, args.output_dir)

    return


# class Creator:
#     """
#     This object can be instantiated with an output directory, *outdir*, and optional
#     *session* and *template_path* directories, which the object maintains.
#
#     This object can simply be called which results in a raw product TIFF file and JSON
#     file being created, written to disk and possibly added to the database.
#
#     This is basically a persistent version of the create() function, so that a
#     database connection can be kept alive (during a Yamcs subscription, for example),
#     and just called with new data.
#
#     All of the arguments to initialize the object are optional:  If *outdir* is not
#     given, the current working directory will be used (beware!).  If *session* is not
#     given, no writes to a database will occur.
#     """
#
#     def __init__(
#         self,
#         outdir: Path = Path.cwd(),
#         session: Optional[Session] = None,
#     ):
#         self.outdir = outdir
#         self.session = session
#
#     def __call__(self, metadata: dict, image: Union[ImageType, Path, None] = None):
#         rp = make_raw_product(metadata, image, self.outdir)
#         logger.info(f"{rp.product_id} created.")
#
#         write_json(rp.asdict(), self.outdir)
#
#         if self.session is not None:
#             with self.session.begin() as s:
#                 s.add(rp)
#
#         return rp
#
#     def from_yamcs_parameters(self, data):
#         for parameter in data.parameters:
#             logger.info(f"{parameter.generation_time} - {parameter.name}")
#             # These are hard-coded until we figure out where they come from.
#             d = {
#                 "bad_pixel_table_id": 0,
#                 "hazlight_aft_port_on": False,
#                 "hazlight_aft_starboard_on": False,
#                 "hazlight_center_port_on": False,
#                 "hazlight_center_starboard_on": False,
#                 "hazlight_fore_port_on": False,
#                 "hazlight_fore_starboard_on": False,
#                 "navlight_left_on": False,
#                 "navlight_right_on": False,
#                 "mission_phase": "TEST",
#                 "purpose": "Navigation",
#             }
#             d.update(parameter.eng_value["imageHeader"])
#             d["yamcs_name"] = parameter.name
#             d["yamcs_generation_time"] = parameter.generation_time
#
#             with io.BytesIO(parameter.eng_value["imageData"]) as f:
#                 im = imread(f)
#
#             self.__call__(d, im)


def get_lights(ir: ImageRecord, session: Union[Session, None] = None):
    # If session is given, values in the ImageRecord are ignored.
    lights = {k: False for k in luminaire_names.values()}
    for short in luminaire_shortnames:
        light_name = luminaire_names[luminaire_shortnames[short]]
        light_col = getattr(ir, f"light_on_{short}")
        if session is not None:
            prev_stmt = (
                select(LightRecord)
                .where(
                    and_(
                        LightRecord.name == light_name,
                        LightRecord.datetime < ir.start_time,
                    )
                )
                .order_by(LightRecord.datetime.desc())
            )
            prev_light = session.scalars(prev_stmt).first()

            if (
                prev_light is not None
                and prev_light.on
                and ir.start_time - prev_light.datetime < timedelta(seconds=10)
            ):
                lights[light_name] = True
        elif light_col is not None:
            lights[light_name] = light_col

    return lights


def label_dict(ir: ImageRecord, lights: dict):
    """Returns a dictionary suitable for label generation."""
    # _inst = ir.instrument_name.lower().replace(" ", "_")
    onoff = {True: "On", False: "Off", None: None}
    pid = pds.VISID(ir.product_id)
    d = dict(
        data_quality="",
        lid=f"{lids['bundle']}:data_raw:{ir.product_id}",
        mission_lid=lids["mission"],
        sc_lid=lids["spacecraft"],
        inst_lid=f"{lids['instrument']}",
        gain_number=(ir.adc_gain * ir.pga_gain),
        exposure_type="Auto" if ir.auto_exposure else "Manual",
        image_filters="",
        led_wavelength=453,  # nm
        luminaires={},
        compression_class=pid.compression_class(),
        observational_intent={},
        onboard_compression_ratio=ir.icer_byte_quota / (2048 * 2048 * 2),
        onboard_compression_type="ICER",
        sample_bits=12,
        sample_bit_mask="2#0000111111111111",
    )
    for k, v in lights.items():
        d["luminaires"][k] = onoff[v]

    try:
        proc_info = ProcessingStage(ir.processing_info)
    except ValueError:
        # processing_info is some bad yamcs value, for now:
        proc_info = ProcessingStage.FLATFIELD | ProcessingStage.LINEARIZATION
        if pid.compression == "s":
            proc_info |= ProcessingStage.SLOG
        warn(
            f"processing_info ({ir.processing_info}) is not one "
            f"of {list(ProcessingStage)}, so assuming a value of {proc_info}"
        )

    im_filt = list()
    if ProcessingStage.FLATFIELD in proc_info:
        im_filt.append("Flat field normalization.")

    if ProcessingStage.LINEARIZATION in proc_info:
        im_filt.append("Linearization.")

    if ProcessingStage.SLOG in proc_info:
        im_filt.append("Sign of the Laplacian of the Gaussian, SLoG.")
        d["sample_bits"] = 8
        d["sample_bit_mask"] = "2#11111111"

    if len(im_filt) > 0:
        d["image_filters"] = " ".join(im_filt)

    if ir.image_request is not None:
        d["observational_intent"]["goal"] = ir.image_request.justification
        d["observational_intent"]["task"] = ir.image_request.title
        d["observational_intent"][
            "activity_id"
        ] = f"Image Request {ir.image_request.id}"
        d["observational_intent"]["target_id"] = ir.image_request.target_location

    if ir.verified is not None:
        if ir.verified:
            d["data_quality"] += "Image manually verified."
        else:
            d["data_quality"] += "Image determined to have errors."

        if ir.verification_notes is not None:
            d["data_quality"] += " " + ir.verification_notes

    return d



#!/usr/bin/env python
# coding: utf-8

"""Defines the VIS image_stats table using the SQLAlchemy ORM."""

# Copyright 2022-2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from sqlalchemy import (
    Float,
    ForeignKey,
    Identity,
    Integer,
    String,
)
from sqlalchemy.orm import backref, mapped_column, relationship

from vipersci.vis.db import Base
from vipersci.vis.db.image_records import ImageRecord


class ImageStats(Base):
    """An object to represent rows in the raw_stats table for VIS."""

    __tablename__ = "image_stats"

    id = mapped_column(Integer, Identity(start=1), primary_key=True)
    product_id = mapped_column(
        String, ForeignKey(ImageRecord.product_id), nullable=False
    )
    raw_product = relationship(ImageRecord, backref=backref("stats", uselist=False))

    blur = mapped_column(
        Float, nullable=False, doc="Blur metric from skimage.measure.blur_effect()."
    )
    mean = mapped_column(Float, nullable=False, doc="Image mean.")
    std = mapped_column(Float, nullable=False, doc="Image standard deviation.")



"""Helps to build PDS4 Bundle and Collection XML labels and files.
"""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import csv
import logging
from pathlib import Path
from typing import List, Dict, Optional
import xml.etree.ElementTree as ET

import pandas as pd

from vipersci.pds.datetime import fromisozformat, isozformat
from vipersci.pds.xml import find_text, ns
from genshi.template import MarkupTemplate

logger = logging.getLogger(__name__)


def get_common_label_info(element: ET.Element, area="pds:Observation_Area"):
    """
    Returns a dict of information harvested from *element* which contains
    PDS4 label information.

    The *area* argument contains the XML element which contains elements
    like start_date_time.  This is likely either 'pds:Observation_Area' (the
    default) or 'pds:Context_Area'.
    """
    if area is not None:
        osc = element.find(".//pds:Observing_System_Component[pds:type='Host']", ns)

        host_name = find_text(osc, "pds:name")
        host_lid = find_text(osc, "pds:Internal_Reference/pds:lid_reference")

        instruments = {}
        for i in element.findall(
            ".//pds:Observing_System_Component[pds:type='Instrument']", ns
        ):
            instruments[find_text(i, "pds:name")] = find_text(
                i, "pds:Internal_Reference/pds:lid_reference"
            )

        purposes = []
        for p in element.findall(
            f"./{area}/pds:Primary_Result_Summary/pds:purpose", ns
        ):
            purposes.append(p.text)

        processing_levels = []
        for p in element.findall(
            f"./{area}/pds:Primary_Result_Summary/pds:processing_level", ns
        ):
            processing_levels.append(p.text)
    else:
        host_name = None
        host_lid = None
        instruments = None
        purposes = None
        processing_levels = None

    lid = find_text(element, "./pds:Identification_Area/pds:logical_identifier")
    d = {
        "lid": lid,
        "vid": find_text(element, "./pds:Identification_Area/pds:version_id"),
        "start_date_time": (
            None
            if area is None
            else fromisozformat(
                find_text(element, f"./{area}/pds:Time_Coordinates/pds:start_date_time")
            )
        ),
        "stop_date_time": (
            None
            if area is None
            else fromisozformat(
                find_text(element, f"./{area}/pds:Time_Coordinates/pds:stop_date_time")
            )
        ),
        "investigation_name": (
            None
            if area is None
            else find_text(element, f"./{area}/pds:Investigation_Area/pds:name")
        ),
        "investigation_type": (
            None
            if area is None
            else find_text(element, f"./{area}/pds:Investigation_Area/pds:type")
        ),
        "investigation_lid": (
            None
            if area is None
            else find_text(
                element,
                f"./{area}/pds:Investigation_Area/pds:Internal_Reference/"
                "pds:lid_reference",
            )
        ),
        "host_name": host_name,
        "host_lid": host_lid,
        "target_name": (
            None
            if area is None
            else find_text(element, ".//pds:Target_Identification/pds:name")
        ),
        "target_type": (
            None
            if area is None
            else find_text(element, ".//pds:Target_Identification/pds:type")
        ),
        "target_lid": (
            None
            if area is None
            else find_text(
                element,
                ".//pds:Target_Identification/pds:Internal_Reference/pds:lid_reference",
            )
        ),
        "instruments": instruments,
        "purposes": purposes,
        "processing_levels": processing_levels,
    }

    return d


def get_lidvid(element: ET.Element):
    return {
        "lid": find_text(element, "./pds:Identification_Area/pds:logical_identifier"),
        "vid": find_text(element, "./pds:Identification_Area/pds:version_id"),
    }


def get_lidvidfile(path: Path) -> dict:
    """
    Returns a dict with three keys: 'lid', 'vid', and 'productfile' whose values
    are extracted from the XML file at *path*.

    This is a convenience function to read a PDS4 XML file and extract the following "
    values:
    ./pds:Identification_Area/pds:logical_identifier,
    ./pds:Identification_Area/pds:version_id, and
    ./pds:File_Area_Observational/pds:File/pds:file_name .
    """
    logger.info(f"Reading {path}")

    root = ET.fromstring(path.read_text())

    d = get_lidvid(root)
    for fxpath in (
        "./pds:File_Area_Observational/pds:File/pds:file_name",
        "./pds:Document/pds:Document_Edition/pds:Document_File/pds:file_name",
        "./pds:File_Area_Browse/pds:File/pds:file_name",
    ):
        element = root.find(fxpath, ns)
        if element is not None:
            el_text = element.text
            if el_text:
                d["productfile"] = el_text
                break
    else:
        d["productfile"] = None

    return d


def gather_info(df, modification_details: List[Dict]):
    """
    Returns a dict of information harvested from the provided pandas dataframe, *df*.

    The dataframe is primarily derived from a list of dicts produced by the
    get_common_label_info() function.
    """
    d = {
        "vid": str(vid_max(modification_details, pd.to_numeric(df["vid"]).max())),
        "instruments": {},
        "purposes": set(),
        "processing_levels": set(),
        "start_date_time": isozformat(df["start_date_time"].min()),
        "stop_date_time": isozformat(df["stop_date_time"].max()),
    }

    for inst_dict in df["instruments"].dropna().tolist():
        d["instruments"].update(inst_dict)

    for p in df["purposes"].dropna().tolist():
        d["purposes"].update(set(p))

    for pl in df["processing_levels"].dropna().tolist():
        d["processing_levels"].update(set(pl))

    return d


def assert_unique(value, series: pd.Series):
    """
    If the *series* has only a single unique value and that value is identical
    to *value* then this function silently returns.  Otherwise it will raise
    a ValueError.

    If the *series* has more than one unique value, this function will raise a
    ValueError.  If the *series* has a single unique value, but it is not equivalent
    to *value*, this function will raise a ValueError.
    """
    s = series.dropna()
    if s.nunique() == 1:
        col_val = s.unique()[0]
        if value != col_val:
            raise ValueError(
                f"The series has a unique value ({col_val}) which does not match the "
                f"provided value ({value}) for {series.name}."
            )
    else:
        raise ValueError(
            f"The series has more than one unique value of {series.name}: "
            f"{series.unique()} "
        )
    return


def vid_max(modification_details: List[Dict], max_product_vid: Optional[float] = None):
    """
    Returns the maximum VID from the *modification_details* list, and if
    *max_product_vid* is not None, then this maximum VID is checked to at least
    be greater than *max_product_vid*, otherwise a ValueError is raised.
    """
    mod_vids = []
    for detail in modification_details:
        mod_vids.append(float(detail["version"]))
    max_mod_vid = max(mod_vids)
    if max_product_vid is not None:
        if max_mod_vid < max_product_vid:
            raise ValueError(
                f"The largest version in the configuration file is {max_mod_vid} but "
                f"the largest version amongst the labels is {max_product_vid} ."
            )
    return max_mod_vid


def write_inventory(path: Path, labels: List[Dict], member="P"):
    """
    Writes a PDS4 collection inventory CSV file to *path* based on the list of
    dicts in *labels*.

    Those dictionaries should have a 'lid' and a 'vid' keyword. The value of *member*
    indicates whether the label is a Primary (P) member or a Secondary (S) member of
    the collection.  If the dictionaries in *labels* have a "member" key, that value
    will be used instead of *member* for that output row.
    """
    with open(path, "w", newline="") as csvfile:
        writer = csv.writer(csvfile)

        for label in labels:
            lidvid = label["lid"] + "::" + label["vid"]
            m = label.get("member", member)
            writer.writerow([m, lidvid])

    return


def write_xml(metadata: dict, outpath: Path, template_path: Path):
    """
    Convenience function to write an XML file at *outpath* that is the
    result of taking the information in *metadata* and populating the XML
    template at *template_path* with that information.

    This function uses the genshi markup library: https://genshi.edgewall.org
    """
    tmpl = MarkupTemplate(template_path.read_text())
    logger.debug(metadata)

    stream = tmpl.generate(**metadata)
    outpath.write_text(stream.render())
    return



#!/usr/bin/env python
"""This module has tests for the validators module."""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from datetime import datetime, timezone, timedelta
import unittest

from vipersci.pds.datetime import fromisozformat
import vipersci.vis.db.validators as vld


class TestValidators(unittest.TestCase):
    def test_validate_datetime_asutc(self):
        dt = datetime(2022, 1, 27, 0, 0, 0, tzinfo=timezone.utc)
        valid = vld.validate_datetime_asutc("foo", dt)
        self.assertEqual(valid, dt)

        t_str = "2023-11-25T14:38:59Z"
        valid = vld.validate_datetime_asutc("foo", t_str)
        self.assertEqual(valid, fromisozformat(t_str))

        # t2_str = "2023-11-25T14:38:59"
        # valid = vld.validate_datetime_asutc("foo", t2_str)
        # self.assertEqual(valid, datetime.fromisoformat(t2_str) )

        dt_e = datetime(2022, 1, 27, 0, 0, 0)
        self.assertRaises(ValueError, vld.validate_datetime_asutc, "foo", dt_e)

        dt_offset = datetime(2022, 1, 27, 0, 0, 0, tzinfo=timezone(timedelta(hours=2)))
        self.assertRaises(ValueError, vld.validate_datetime_asutc, "foo", dt_offset)

        self.assertRaises(
            ValueError,
            vld.validate_datetime_asutc,
            "foo",
            "not interpretable as a datetime",
        )

    def test_validate_purpose(self):
        self.assertEqual(vld.validate_purpose("Science"), "SCIENCE")

        self.assertRaises(ValueError, vld.validate_purpose, "not a Purpose")



"""
The bounds module is used to calculate a bounding box for a set of 2d spatial data.
"""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import math
from typing import Callable, Tuple

import numpy as np
from numpy.typing import NDArray


def compute_bounds(
    x_arr: NDArray,
    y_arr: NDArray,
) -> Tuple[float, float, float, float]:
    """Compute cartesian bounds of input coordinates

    Args:
        x_arr: set of x locations
        y_arr: set of y locations

    Returns:
        a tuple (left, bottom, right, top) describing the bounds of the input data
    """
    return (np.amin(x_arr), np.amin(y_arr), np.amax(x_arr), np.amax(y_arr))


def pad_grid_align_bounds(
    left: float,
    bottom: float,
    right: float,
    top: float,
    ground_sample_distance: float,
    padding: int = 0,
) -> Tuple[float, float, float, float]:
    """Given bounds, add a padding and align to a grid of size of
    ground_sample_distance.  That is, the output bounds will be integer multiples of
    ground_sample_distance.

    Args:
        left: xmin of input bounds
        bottom: ymin of input bounds
        right: xmax of input bounds
        top: ymax of input bounds
        ground_sample_distance: Spatial resolution of grid
        padding: Pixels (multiples of ground_sample_distance) of padding to add on
          to bounds
    Returns:
        a tuple (left, bottom, right, top) describing padded and aligned bounds
    """

    def _grid_snap(
        operator: Callable[[float], float],
        value: float,
        ground_sample_distance: float,
    ) -> float:
        """Round value to closest multiple of ground_sample_distance.

        Args:
            operator: Function that defines how value is rounded.  Usually math.floor
              or math.ceil.
            value: value to snap to grid
            ground_sample_distance: defines a grid with origin 0 and cells of this size.

        Returns:
            Rounded value
        """
        return operator(value / ground_sample_distance) * ground_sample_distance

    # Convert from pixels of padding to a buffer in x/y distance units.
    buffer = padding * ground_sample_distance
    # Make sure that bounds are snapped into a grid centered on the
    # origin with a *ground_sample_distance* step.
    left = _grid_snap(math.floor, left - buffer, ground_sample_distance)
    right = _grid_snap(math.ceil, right + buffer, ground_sample_distance)
    bottom = _grid_snap(math.floor, bottom - buffer, ground_sample_distance)
    top = _grid_snap(math.ceil, top + buffer, ground_sample_distance)

    return (left, bottom, right, top)



"""Helps to build PDS4 Bundle and Collection XML labels and files.
"""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse

from vipersci.pds.labelmaker.bundle import add_parser as bundle_ap
from vipersci.pds.labelmaker.collection import add_parser as collection_ap
from vipersci.pds.labelmaker.generic import add_parser as generic_ap
from vipersci.pds.labelmaker.inventory import add_parser as inventory_ap
from vipersci import util


def main():
    parser = argparse.ArgumentParser(
        description=__doc__, parents=[util.parent_parser()]
    )
    subparsers = parser.add_subparsers(
        title="subcommands",
        description="valid subcommands",
        help="additional help available via '%(prog)s subcommand -h'",
    )
    bundle_ap(subparsers)
    collection_ap(subparsers)
    inventory_ap(subparsers)
    generic_ap(subparsers)
    args = parser.parse_args()
    args.func(args)

    return
